/*
 * Copyright (c) 2016, Alliance for Open Media. All rights reserved
 *
 * This source code is subject to the terms of the BSD 2 Clause License and
 * the Alliance for Open Media Patent License 1.0. If the BSD 2 Clause License
 * was not distributed with this source code in the LICENSE file, you can
 * obtain it at www.aomedia.org/license/software. If the Alliance for Open
 * Media Patent License 1.0 was not distributed with this source code in the
 * PATENTS file, you can obtain it at www.aomedia.org/license/patent.
 */

#include "./aom_config.h"
#include "av1/common/entropy.h"
#include "av1/common/blockd.h"
#include "av1/common/onyxc_int.h"
#include "av1/common/entropymode.h"
#include "av1/common/scan.h"
#include "aom_mem/aom_mem.h"
#include "aom/aom_integer.h"

// Unconstrained Node Tree
/* clang-format off */
const aom_tree_index av1_coef_con_tree[TREE_SIZE(ENTROPY_TOKENS)] = {
  2, 6,                                // 0 = LOW_VAL
  -TWO_TOKEN, 4,                       // 1 = TWO
  -THREE_TOKEN, -FOUR_TOKEN,           // 2 = THREE
  8, 10,                               // 3 = HIGH_LOW
  -CATEGORY1_TOKEN, -CATEGORY2_TOKEN,  // 4 = CAT_ONE
  12, 14,                              // 5 = CAT_THREEFOUR
  -CATEGORY3_TOKEN, -CATEGORY4_TOKEN,  // 6 = CAT_THREE
  -CATEGORY5_TOKEN, -CATEGORY6_TOKEN   // 7 = CAT_FIVE
};
/* clang-format on */

const aom_prob av1_cat1_prob[] = { 159 };
const aom_prob av1_cat2_prob[] = { 165, 145 };
const aom_prob av1_cat3_prob[] = { 173, 148, 140 };
const aom_prob av1_cat4_prob[] = { 176, 155, 140, 135 };
const aom_prob av1_cat5_prob[] = { 180, 157, 141, 134, 130 };
const aom_prob av1_cat6_prob[] = { 254, 254, 254, 252, 249, 243, 230,
                                   196, 177, 153, 140, 133, 130, 129 };
#if CONFIG_AOM_HIGHBITDEPTH
const aom_prob av1_cat1_prob_high10[] = { 159 };
const aom_prob av1_cat2_prob_high10[] = { 165, 145 };
const aom_prob av1_cat3_prob_high10[] = { 173, 148, 140 };
const aom_prob av1_cat4_prob_high10[] = { 176, 155, 140, 135 };
const aom_prob av1_cat5_prob_high10[] = { 180, 157, 141, 134, 130 };
const aom_prob av1_cat6_prob_high10[] = {
  255, 255, 254, 254, 254, 252, 249, 243, 230, 196, 177, 153, 140, 133, 130, 129
};
const aom_prob av1_cat1_prob_high12[] = { 159 };
const aom_prob av1_cat2_prob_high12[] = { 165, 145 };
const aom_prob av1_cat3_prob_high12[] = { 173, 148, 140 };
const aom_prob av1_cat4_prob_high12[] = { 176, 155, 140, 135 };
const aom_prob av1_cat5_prob_high12[] = { 180, 157, 141, 134, 130 };
const aom_prob av1_cat6_prob_high12[] = { 255, 255, 255, 255, 254, 254,
                                          254, 252, 249, 243, 230, 196,
                                          177, 153, 140, 133, 130, 129 };
#endif

const uint16_t band_count_table[TX_SIZES_ALL][8] = {
#if CONFIG_CB4X4
  { 1, 2, 2, 3, 0, 0, 0 },
#endif
  { 1, 2, 3, 4, 3, 16 - 13, 0 },    { 1, 2, 3, 4, 11, 64 - 21, 0 },
  { 1, 2, 3, 4, 11, 256 - 21, 0 },  { 1, 2, 3, 4, 11, 1024 - 21, 0 },
#if CONFIG_TX64X64
  { 1, 2, 3, 4, 11, 4096 - 21, 0 },
#endif  // CONFIG_TX64X64
  { 1, 2, 3, 4, 8, 32 - 18, 0 },    { 1, 2, 3, 4, 8, 32 - 18, 0 },
  { 1, 2, 3, 4, 11, 128 - 21, 0 },  { 1, 2, 3, 4, 11, 128 - 21, 0 },
  { 1, 2, 3, 4, 11, 512 - 21, 0 },  { 1, 2, 3, 4, 11, 512 - 21, 0 },
  { 1, 2, 3, 4, 11, 64 - 21, 0 },   { 1, 2, 3, 4, 11, 64 - 21, 0 },
  { 1, 2, 3, 4, 11, 256 - 21, 0 },  { 1, 2, 3, 4, 11, 256 - 21, 0 },
};

const uint16_t band_cum_count_table[TX_SIZES_ALL][8] = {
#if CONFIG_CB4X4
  { 0, 1, 3, 6, 10, 13, 16, 0 },
#endif
  { 0, 1, 3, 6, 10, 13, 16, 0 },   { 0, 1, 3, 6, 10, 21, 64, 0 },
  { 0, 1, 3, 6, 10, 21, 256, 0 },  { 0, 1, 3, 6, 10, 21, 1024, 0 },
#if CONFIG_TX64X64
  { 0, 1, 3, 6, 10, 21, 4096, 0 },
#endif  // CONFIG_TX64X64
  { 0, 1, 3, 6, 10, 18, 32, 0 },   { 0, 1, 3, 6, 10, 18, 32, 0 },
  { 0, 1, 3, 6, 10, 21, 128, 0 },  { 0, 1, 3, 6, 10, 21, 128, 0 },
  { 0, 1, 3, 6, 10, 21, 512, 0 },  { 0, 1, 3, 6, 10, 21, 512, 0 },
  { 0, 1, 3, 6, 10, 21, 64, 0 },   { 0, 1, 3, 6, 10, 21, 64, 0 },
  { 0, 1, 3, 6, 10, 21, 256, 0 },  { 0, 1, 3, 6, 10, 21, 256, 0 },
};

const uint8_t av1_coefband_trans_8x8plus[MAX_TX_SQUARE] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5,
  // beyond MAXBAND_INDEX+1 all values are filled as 5
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
#if CONFIG_TX64X64
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
  5, 5, 5, 5
#endif  // CONFIG_TX64X64
};

const uint8_t av1_coefband_trans_4x8_8x4[32] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4,
  4, 4, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
};

const uint8_t av1_coefband_trans_4x4[16] = {
  0, 1, 1, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 5, 5, 5,
};

const uint8_t av1_pt_energy_class[ENTROPY_TOKENS] = { 0, 1, 2, 3, 3, 4,
                                                      4, 5, 5, 5, 5, 5 };

// Model obtained from a 2-sided zero-centered distribution derived
// from a Pareto distribution. The cdf of the distribution is:
// cdf(x) = 0.5 + 0.5 * sgn(x) * [1 - {alpha/(alpha + |x|)} ^ beta]
//
// For a given beta and a given probablity of the 1-node, the alpha
// is first solved, and then the {alpha, beta} pair is used to generate
// the probabilities for the rest of the nodes.

// beta = 8

// Every odd line in this table can be generated from the even lines
// by averaging :
// av1_pareto8_full[l][node] = (av1_pareto8_full[l-1][node] +
//                              av1_pareto8_full[l+1][node] ) >> 1;
const aom_prob av1_pareto8_full[COEFF_PROB_MODELS][MODEL_NODES] = {
  { 3, 86, 128, 6, 86, 23, 88, 29 },
  { 6, 86, 128, 11, 87, 42, 91, 52 },
  { 9, 86, 129, 17, 88, 61, 94, 76 },
  { 12, 86, 129, 22, 88, 77, 97, 93 },
  { 15, 87, 129, 28, 89, 93, 100, 110 },
  { 17, 87, 129, 33, 90, 105, 103, 123 },
  { 20, 88, 130, 38, 91, 118, 106, 136 },
  { 23, 88, 130, 43, 91, 128, 108, 146 },
  { 26, 89, 131, 48, 92, 139, 111, 156 },
  { 28, 89, 131, 53, 93, 147, 114, 163 },
  { 31, 90, 131, 58, 94, 156, 117, 171 },
  { 34, 90, 131, 62, 94, 163, 119, 177 },
  { 37, 90, 132, 66, 95, 171, 122, 184 },
  { 39, 90, 132, 70, 96, 177, 124, 189 },
  { 42, 91, 132, 75, 97, 183, 127, 194 },
  { 44, 91, 132, 79, 97, 188, 129, 198 },
  { 47, 92, 133, 83, 98, 193, 132, 202 },
  { 49, 92, 133, 86, 99, 197, 134, 205 },
  { 52, 93, 133, 90, 100, 201, 137, 208 },
  { 54, 93, 133, 94, 100, 204, 139, 211 },
  { 57, 94, 134, 98, 101, 208, 142, 214 },
  { 59, 94, 134, 101, 102, 211, 144, 216 },
  { 62, 94, 135, 105, 103, 214, 146, 218 },
  { 64, 94, 135, 108, 103, 216, 148, 220 },
  { 66, 95, 135, 111, 104, 219, 151, 222 },
  { 68, 95, 135, 114, 105, 221, 153, 223 },
  { 71, 96, 136, 117, 106, 224, 155, 225 },
  { 73, 96, 136, 120, 106, 225, 157, 226 },
  { 76, 97, 136, 123, 107, 227, 159, 228 },
  { 78, 97, 136, 126, 108, 229, 160, 229 },
  { 80, 98, 137, 129, 109, 231, 162, 231 },
  { 82, 98, 137, 131, 109, 232, 164, 232 },
  { 84, 98, 138, 134, 110, 234, 166, 233 },
  { 86, 98, 138, 137, 111, 235, 168, 234 },
  { 89, 99, 138, 140, 112, 236, 170, 235 },
  { 91, 99, 138, 142, 112, 237, 171, 235 },
  { 93, 100, 139, 145, 113, 238, 173, 236 },
  { 95, 100, 139, 147, 114, 239, 174, 237 },
  { 97, 101, 140, 149, 115, 240, 176, 238 },
  { 99, 101, 140, 151, 115, 241, 177, 238 },
  { 101, 102, 140, 154, 116, 242, 179, 239 },
  { 103, 102, 140, 156, 117, 242, 180, 239 },
  { 105, 103, 141, 158, 118, 243, 182, 240 },
  { 107, 103, 141, 160, 118, 243, 183, 240 },
  { 109, 104, 141, 162, 119, 244, 185, 241 },
  { 111, 104, 141, 164, 119, 244, 186, 241 },
  { 113, 104, 142, 166, 120, 245, 187, 242 },
  { 114, 104, 142, 168, 121, 245, 188, 242 },
  { 116, 105, 143, 170, 122, 246, 190, 243 },
  { 118, 105, 143, 171, 122, 246, 191, 243 },
  { 120, 106, 143, 173, 123, 247, 192, 244 },
  { 121, 106, 143, 175, 124, 247, 193, 244 },
  { 123, 107, 144, 177, 125, 248, 195, 244 },
  { 125, 107, 144, 178, 125, 248, 196, 244 },
  { 127, 108, 145, 180, 126, 249, 197, 245 },
  { 128, 108, 145, 181, 127, 249, 198, 245 },
  { 130, 109, 145, 183, 128, 249, 199, 245 },
  { 132, 109, 145, 184, 128, 249, 200, 245 },
  { 134, 110, 146, 186, 129, 250, 201, 246 },
  { 135, 110, 146, 187, 130, 250, 202, 246 },
  { 137, 111, 147, 189, 131, 251, 203, 246 },
  { 138, 111, 147, 190, 131, 251, 204, 246 },
  { 140, 112, 147, 192, 132, 251, 205, 247 },
  { 141, 112, 147, 193, 132, 251, 206, 247 },
  { 143, 113, 148, 194, 133, 251, 207, 247 },
  { 144, 113, 148, 195, 134, 251, 207, 247 },
  { 146, 114, 149, 197, 135, 252, 208, 248 },
  { 147, 114, 149, 198, 135, 252, 209, 248 },
  { 149, 115, 149, 199, 136, 252, 210, 248 },
  { 150, 115, 149, 200, 137, 252, 210, 248 },
  { 152, 115, 150, 201, 138, 252, 211, 248 },
  { 153, 115, 150, 202, 138, 252, 212, 248 },
  { 155, 116, 151, 204, 139, 253, 213, 249 },
  { 156, 116, 151, 205, 139, 253, 213, 249 },
  { 158, 117, 151, 206, 140, 253, 214, 249 },
  { 159, 117, 151, 207, 141, 253, 215, 249 },
  { 161, 118, 152, 208, 142, 253, 216, 249 },
  { 162, 118, 152, 209, 142, 253, 216, 249 },
  { 163, 119, 153, 210, 143, 253, 217, 249 },
  { 164, 119, 153, 211, 143, 253, 217, 249 },
  { 166, 120, 153, 212, 144, 254, 218, 250 },
  { 167, 120, 153, 212, 145, 254, 219, 250 },
  { 168, 121, 154, 213, 146, 254, 220, 250 },
  { 169, 121, 154, 214, 146, 254, 220, 250 },
  { 171, 122, 155, 215, 147, 254, 221, 250 },
  { 172, 122, 155, 216, 147, 254, 221, 250 },
  { 173, 123, 155, 217, 148, 254, 222, 250 },
  { 174, 123, 155, 217, 149, 254, 222, 250 },
  { 176, 124, 156, 218, 150, 254, 223, 250 },
  { 177, 124, 156, 219, 150, 254, 223, 250 },
  { 178, 125, 157, 220, 151, 254, 224, 251 },
  { 179, 125, 157, 220, 151, 254, 224, 251 },
  { 180, 126, 157, 221, 152, 254, 225, 251 },
  { 181, 126, 157, 221, 152, 254, 225, 251 },
  { 183, 127, 158, 222, 153, 254, 226, 251 },
  { 184, 127, 158, 223, 154, 254, 226, 251 },
  { 185, 128, 159, 224, 155, 255, 227, 251 },
  { 186, 128, 159, 224, 155, 255, 227, 251 },
  { 187, 129, 160, 225, 156, 255, 228, 251 },
  { 188, 130, 160, 225, 156, 255, 228, 251 },
  { 189, 131, 160, 226, 157, 255, 228, 251 },
  { 190, 131, 160, 226, 158, 255, 228, 251 },
  { 191, 132, 161, 227, 159, 255, 229, 251 },
  { 192, 132, 161, 227, 159, 255, 229, 251 },
  { 193, 133, 162, 228, 160, 255, 230, 252 },
  { 194, 133, 162, 229, 160, 255, 230, 252 },
  { 195, 134, 163, 230, 161, 255, 231, 252 },
  { 196, 134, 163, 230, 161, 255, 231, 252 },
  { 197, 135, 163, 231, 162, 255, 231, 252 },
  { 198, 135, 163, 231, 162, 255, 231, 252 },
  { 199, 136, 164, 232, 163, 255, 232, 252 },
  { 200, 136, 164, 232, 164, 255, 232, 252 },
  { 201, 137, 165, 233, 165, 255, 233, 252 },
  { 201, 137, 165, 233, 165, 255, 233, 252 },
  { 202, 138, 166, 233, 166, 255, 233, 252 },
  { 203, 138, 166, 233, 166, 255, 233, 252 },
  { 204, 139, 166, 234, 167, 255, 234, 252 },
  { 205, 139, 166, 234, 167, 255, 234, 252 },
  { 206, 140, 167, 235, 168, 255, 235, 252 },
  { 206, 140, 167, 235, 168, 255, 235, 252 },
  { 207, 141, 168, 236, 169, 255, 235, 252 },
  { 208, 141, 168, 236, 170, 255, 235, 252 },
  { 209, 142, 169, 237, 171, 255, 236, 252 },
  { 209, 143, 169, 237, 171, 255, 236, 252 },
  { 210, 144, 169, 237, 172, 255, 236, 252 },
  { 211, 144, 169, 237, 172, 255, 236, 252 },
  { 212, 145, 170, 238, 173, 255, 237, 252 },
  { 213, 145, 170, 238, 173, 255, 237, 252 },
  { 214, 146, 171, 239, 174, 255, 237, 253 },
  { 214, 146, 171, 239, 174, 255, 237, 253 },
  { 215, 147, 172, 240, 175, 255, 238, 253 },
  { 215, 147, 172, 240, 175, 255, 238, 253 },
  { 216, 148, 173, 240, 176, 255, 238, 253 },
  { 217, 148, 173, 240, 176, 255, 238, 253 },
  { 218, 149, 173, 241, 177, 255, 239, 253 },
  { 218, 149, 173, 241, 178, 255, 239, 253 },
  { 219, 150, 174, 241, 179, 255, 239, 253 },
  { 219, 151, 174, 241, 179, 255, 239, 253 },
  { 220, 152, 175, 242, 180, 255, 240, 253 },
  { 221, 152, 175, 242, 180, 255, 240, 253 },
  { 222, 153, 176, 242, 181, 255, 240, 253 },
  { 222, 153, 176, 242, 181, 255, 240, 253 },
  { 223, 154, 177, 243, 182, 255, 240, 253 },
  { 223, 154, 177, 243, 182, 255, 240, 253 },
  { 224, 155, 178, 244, 183, 255, 241, 253 },
  { 224, 155, 178, 244, 183, 255, 241, 253 },
  { 225, 156, 178, 244, 184, 255, 241, 253 },
  { 225, 157, 178, 244, 184, 255, 241, 253 },
  { 226, 158, 179, 244, 185, 255, 242, 253 },
  { 227, 158, 179, 244, 185, 255, 242, 253 },
  { 228, 159, 180, 245, 186, 255, 242, 253 },
  { 228, 159, 180, 245, 186, 255, 242, 253 },
  { 229, 160, 181, 245, 187, 255, 242, 253 },
  { 229, 160, 181, 245, 187, 255, 242, 253 },
  { 230, 161, 182, 246, 188, 255, 243, 253 },
  { 230, 162, 182, 246, 188, 255, 243, 253 },
  { 231, 163, 183, 246, 189, 255, 243, 253 },
  { 231, 163, 183, 246, 189, 255, 243, 253 },
  { 232, 164, 184, 247, 190, 255, 243, 253 },
  { 232, 164, 184, 247, 190, 255, 243, 253 },
  { 233, 165, 185, 247, 191, 255, 244, 253 },
  { 233, 165, 185, 247, 191, 255, 244, 253 },
  { 234, 166, 185, 247, 192, 255, 244, 253 },
  { 234, 167, 185, 247, 192, 255, 244, 253 },
  { 235, 168, 186, 248, 193, 255, 244, 253 },
  { 235, 168, 186, 248, 193, 255, 244, 253 },
  { 236, 169, 187, 248, 194, 255, 244, 253 },
  { 236, 169, 187, 248, 194, 255, 244, 253 },
  { 236, 170, 188, 248, 195, 255, 245, 253 },
  { 236, 170, 188, 248, 195, 255, 245, 253 },
  { 237, 171, 189, 249, 196, 255, 245, 254 },
  { 237, 172, 189, 249, 196, 255, 245, 254 },
  { 238, 173, 190, 249, 197, 255, 245, 254 },
  { 238, 173, 190, 249, 197, 255, 245, 254 },
  { 239, 174, 191, 249, 198, 255, 245, 254 },
  { 239, 174, 191, 249, 198, 255, 245, 254 },
  { 240, 175, 192, 249, 199, 255, 246, 254 },
  { 240, 176, 192, 249, 199, 255, 246, 254 },
  { 240, 177, 193, 250, 200, 255, 246, 254 },
  { 240, 177, 193, 250, 200, 255, 246, 254 },
  { 241, 178, 194, 250, 201, 255, 246, 254 },
  { 241, 178, 194, 250, 201, 255, 246, 254 },
  { 242, 179, 195, 250, 202, 255, 246, 254 },
  { 242, 180, 195, 250, 202, 255, 246, 254 },
  { 242, 181, 196, 250, 203, 255, 247, 254 },
  { 242, 181, 196, 250, 203, 255, 247, 254 },
  { 243, 182, 197, 251, 204, 255, 247, 254 },
  { 243, 183, 197, 251, 204, 255, 247, 254 },
  { 244, 184, 198, 251, 205, 255, 247, 254 },
  { 244, 184, 198, 251, 205, 255, 247, 254 },
  { 244, 185, 199, 251, 206, 255, 247, 254 },
  { 244, 185, 199, 251, 206, 255, 247, 254 },
  { 245, 186, 200, 251, 207, 255, 247, 254 },
  { 245, 187, 200, 251, 207, 255, 247, 254 },
  { 246, 188, 201, 252, 207, 255, 248, 254 },
  { 246, 188, 201, 252, 207, 255, 248, 254 },
  { 246, 189, 202, 252, 208, 255, 248, 254 },
  { 246, 190, 202, 252, 208, 255, 248, 254 },
  { 247, 191, 203, 252, 209, 255, 248, 254 },
  { 247, 191, 203, 252, 209, 255, 248, 254 },
  { 247, 192, 204, 252, 210, 255, 248, 254 },
  { 247, 193, 204, 252, 210, 255, 248, 254 },
  { 248, 194, 205, 252, 211, 255, 248, 254 },
  { 248, 194, 205, 252, 211, 255, 248, 254 },
  { 248, 195, 206, 252, 212, 255, 249, 254 },
  { 248, 196, 206, 252, 212, 255, 249, 254 },
  { 249, 197, 207, 253, 213, 255, 249, 254 },
  { 249, 197, 207, 253, 213, 255, 249, 254 },
  { 249, 198, 208, 253, 214, 255, 249, 254 },
  { 249, 199, 209, 253, 214, 255, 249, 254 },
  { 250, 200, 210, 253, 215, 255, 249, 254 },
  { 250, 200, 210, 253, 215, 255, 249, 254 },
  { 250, 201, 211, 253, 215, 255, 249, 254 },
  { 250, 202, 211, 253, 215, 255, 249, 254 },
  { 250, 203, 212, 253, 216, 255, 249, 254 },
  { 250, 203, 212, 253, 216, 255, 249, 254 },
  { 251, 204, 213, 253, 217, 255, 250, 254 },
  { 251, 205, 213, 253, 217, 255, 250, 254 },
  { 251, 206, 214, 254, 218, 255, 250, 254 },
  { 251, 206, 215, 254, 218, 255, 250, 254 },
  { 252, 207, 216, 254, 219, 255, 250, 254 },
  { 252, 208, 216, 254, 219, 255, 250, 254 },
  { 252, 209, 217, 254, 220, 255, 250, 254 },
  { 252, 210, 217, 254, 220, 255, 250, 254 },
  { 252, 211, 218, 254, 221, 255, 250, 254 },
  { 252, 212, 218, 254, 221, 255, 250, 254 },
  { 253, 213, 219, 254, 222, 255, 250, 254 },
  { 253, 213, 220, 254, 222, 255, 250, 254 },
  { 253, 214, 221, 254, 223, 255, 250, 254 },
  { 253, 215, 221, 254, 223, 255, 250, 254 },
  { 253, 216, 222, 254, 224, 255, 251, 254 },
  { 253, 217, 223, 254, 224, 255, 251, 254 },
  { 253, 218, 224, 254, 225, 255, 251, 254 },
  { 253, 219, 224, 254, 225, 255, 251, 254 },
  { 254, 220, 225, 254, 225, 255, 251, 254 },
  { 254, 221, 226, 254, 225, 255, 251, 254 },
  { 254, 222, 227, 255, 226, 255, 251, 254 },
  { 254, 223, 227, 255, 226, 255, 251, 254 },
  { 254, 224, 228, 255, 227, 255, 251, 254 },
  { 254, 225, 229, 255, 227, 255, 251, 254 },
  { 254, 226, 230, 255, 228, 255, 251, 254 },
  { 254, 227, 230, 255, 229, 255, 251, 254 },
  { 255, 228, 231, 255, 230, 255, 251, 254 },
  { 255, 229, 232, 255, 230, 255, 251, 254 },
  { 255, 230, 233, 255, 231, 255, 252, 254 },
  { 255, 231, 234, 255, 231, 255, 252, 254 },
  { 255, 232, 235, 255, 232, 255, 252, 254 },
  { 255, 233, 236, 255, 232, 255, 252, 254 },
  { 255, 235, 237, 255, 233, 255, 252, 254 },
  { 255, 236, 238, 255, 234, 255, 252, 254 },
  { 255, 238, 240, 255, 235, 255, 252, 255 },
  { 255, 239, 241, 255, 235, 255, 252, 254 },
  { 255, 241, 243, 255, 236, 255, 252, 254 },
  { 255, 243, 245, 255, 237, 255, 252, 254 },
  { 255, 246, 247, 255, 239, 255, 253, 255 },
};

#if CONFIG_EC_MULTISYMBOL
// Model obtained from a 2-sided zero-centered distribution derived
// from a Pareto distribution. The cdf of the distribution is:
// cdf(x) = 0.5 + 0.5 * sgn(x) * [1 - {alpha/(alpha + |x|)} ^ beta]
//
// For a given beta and a given probability of the 1-node, the alpha
// is first solved, and then the {alpha, beta} pair is used to generate
// the probabilities for the rest of the nodes.
//
// The full source code of the generating program is available in:
// tools/gen_constrained_tokenset.py
//
// Values for tokens ONE_TOKEN through CATEGORY6_TOKEN included here.
// ZERO_TOKEN and EOB_TOKEN are coded as flags outside this coder.
const aom_cdf_prob
    av1_pareto8_token_probs[COEFF_PROB_MODELS][ENTROPY_TOKENS - 2] = {
      { 128, 127, 127, 126, 251, 495, 965, 1832, 3305, 25412 },
      { 256, 254, 252, 249, 492, 959, 1820, 3283, 5365, 19838 },
      { 384, 379, 374, 369, 724, 1392, 2574, 4417, 6568, 15587 },
      { 512, 503, 494, 486, 946, 1795, 3238, 5289, 7184, 12321 },
      { 640, 626, 612, 599, 1159, 2172, 3818, 5946, 7401, 9795 },
      { 768, 748, 728, 709, 1363, 2522, 4324, 6424, 7352, 7830 },
      { 896, 869, 842, 816, 1559, 2847, 4762, 6755, 7131, 6291 },
      { 1024, 988, 954, 921, 1747, 3148, 5139, 6966, 6803, 5078 },
      { 1152, 1107, 1063, 1022, 1926, 3427, 5460, 7080, 6412, 4119 },
      { 1280, 1224, 1171, 1120, 2098, 3685, 5730, 7113, 5991, 3356 },
      { 1408, 1340, 1276, 1216, 2261, 3923, 5955, 7083, 5560, 2746 },
      { 1536, 1455, 1380, 1308, 2418, 4142, 6140, 7001, 5133, 2255 },
      { 1664, 1569, 1481, 1398, 2567, 4342, 6287, 6879, 4721, 1860 },
      { 1792, 1683, 1580, 1485, 2709, 4525, 6401, 6725, 4329, 1539 },
      { 1920, 1794, 1678, 1570, 2845, 4692, 6486, 6546, 3959, 1278 },
      { 2048, 1905, 1773, 1651, 2974, 4844, 6543, 6350, 3615, 1065 },
      { 2176, 2015, 1867, 1731, 3096, 4980, 6576, 6140, 3296, 891 },
      { 2304, 2123, 1958, 1807, 3212, 5104, 6589, 5922, 3002, 747 },
      { 2432, 2231, 2048, 1882, 3322, 5214, 6581, 5698, 2732, 628 },
      { 2560, 2337, 2136, 1953, 3427, 5311, 6557, 5472, 2485, 530 },
      { 2688, 2442, 2222, 2023, 3525, 5397, 6518, 5246, 2259, 448 },
      { 2816, 2547, 2306, 2090, 3618, 5472, 6465, 5021, 2053, 380 },
      { 2944, 2650, 2388, 2154, 3706, 5537, 6401, 4799, 1866, 323 },
      { 3072, 2752, 2468, 2217, 3788, 5591, 6327, 4581, 1696, 276 },
      { 3200, 2853, 2547, 2277, 3866, 5637, 6243, 4369, 1541, 235 },
      { 3328, 2952, 2624, 2335, 3938, 5673, 6152, 4163, 1401, 202 },
      { 3456, 3051, 2699, 2391, 4006, 5702, 6054, 3962, 1274, 173 },
      { 3584, 3149, 2772, 2444, 4070, 5723, 5950, 3769, 1158, 149 },
      { 3712, 3246, 2843, 2496, 4128, 5736, 5842, 3583, 1054, 128 },
      { 3840, 3341, 2913, 2545, 4183, 5743, 5729, 3404, 959, 111 },
      { 3968, 3436, 2981, 2593, 4233, 5743, 5614, 3232, 872, 96 },
      { 4096, 3529, 3048, 2638, 4280, 5737, 5496, 3067, 794, 83 },
      { 4224, 3621, 3113, 2682, 4322, 5726, 5375, 2909, 724, 72 },
      { 4352, 3712, 3176, 2724, 4361, 5709, 5253, 2759, 659, 63 },
      { 4480, 3803, 3237, 2764, 4396, 5687, 5130, 2615, 601, 55 },
      { 4608, 3892, 3297, 2801, 4428, 5661, 5007, 2478, 548, 48 },
      { 4736, 3980, 3355, 2838, 4456, 5631, 4883, 2347, 500, 42 },
      { 4864, 4067, 3412, 2872, 4481, 5596, 4760, 2223, 456, 37 },
      { 4992, 4152, 3467, 2905, 4503, 5558, 4637, 2105, 417, 32 },
      { 5120, 4237, 3521, 2936, 4521, 5516, 4515, 1993, 381, 28 },
      { 5248, 4321, 3573, 2966, 4537, 5471, 4393, 1886, 348, 25 },
      { 5376, 4404, 3623, 2993, 4550, 5424, 4273, 1785, 318, 22 },
      { 5504, 4486, 3672, 3020, 4560, 5373, 4155, 1688, 291, 19 },
      { 5632, 4566, 3720, 3044, 4568, 5321, 4037, 1597, 266, 17 },
      { 5760, 4646, 3766, 3067, 4572, 5265, 3922, 1511, 244, 15 },
      { 5888, 4724, 3811, 3089, 4575, 5208, 3808, 1429, 223, 13 },
      { 6016, 4802, 3854, 3109, 4575, 5148, 3696, 1352, 204, 12 },
      { 6144, 4878, 3895, 3128, 4573, 5088, 3587, 1278, 187, 10 },
      { 6272, 4953, 3936, 3145, 4568, 5025, 3479, 1209, 172, 9 },
      { 6400, 5028, 3975, 3161, 4561, 4961, 3373, 1143, 158, 8 },
      { 6528, 5101, 4012, 3175, 4553, 4896, 3270, 1081, 145, 7 },
      { 6656, 5173, 4048, 3189, 4542, 4830, 3168, 1022, 133, 7 },
      { 6784, 5244, 4083, 3201, 4530, 4763, 3069, 966, 122, 6 },
      { 6912, 5314, 4117, 3212, 4516, 4694, 2973, 913, 112, 5 },
      { 7040, 5383, 4149, 3221, 4500, 4626, 2878, 863, 103, 5 },
      { 7168, 5452, 4180, 3229, 4482, 4556, 2786, 816, 95, 4 },
      { 7296, 5519, 4210, 3236, 4463, 4486, 2696, 771, 87, 4 },
      { 7424, 5585, 4238, 3242, 4442, 4416, 2609, 729, 80, 3 },
      { 7552, 5650, 4265, 3247, 4420, 4345, 2523, 689, 74, 3 },
      { 7680, 5714, 4291, 3251, 4396, 4274, 2440, 651, 68, 3 },
      { 7808, 5777, 4315, 3254, 4371, 4203, 2359, 616, 63, 2 },
      { 7936, 5838, 4339, 3255, 4345, 4132, 2281, 582, 58, 2 },
      { 8064, 5899, 4361, 3256, 4318, 4061, 2204, 550, 53, 2 },
      { 8192, 5959, 4382, 3255, 4289, 3990, 2130, 520, 49, 2 },
      { 8320, 6018, 4402, 3254, 4259, 3919, 2057, 492, 45, 2 },
      { 8448, 6075, 4421, 3252, 4229, 3848, 1987, 465, 42, 1 },
      { 8576, 6133, 4438, 3248, 4197, 3778, 1919, 439, 39, 1 },
      { 8704, 6188, 4455, 3244, 4164, 3708, 1853, 415, 36, 1 },
      { 8832, 6243, 4470, 3239, 4131, 3638, 1789, 392, 33, 1 },
      { 8960, 6297, 4484, 3233, 4096, 3569, 1727, 371, 30, 1 },
      { 9088, 6349, 4497, 3226, 4061, 3500, 1667, 351, 28, 1 },
      { 9216, 6401, 4509, 3219, 4025, 3432, 1608, 331, 26, 1 },
      { 9344, 6452, 4520, 3210, 3989, 3364, 1551, 313, 24, 1 },
      { 9472, 6501, 4530, 3201, 3952, 3297, 1496, 296, 22, 1 },
      { 9600, 6550, 4539, 3191, 3914, 3230, 1443, 280, 20, 1 },
      { 9728, 6597, 4547, 3180, 3875, 3164, 1392, 265, 19, 1 },
      { 9856, 6644, 4554, 3169, 3836, 3098, 1342, 250, 18, 1 },
      { 9984, 6690, 4560, 3157, 3796, 3034, 1293, 237, 16, 1 },
      { 10112, 6734, 4565, 3144, 3756, 2970, 1247, 224, 15, 1 },
      { 10240, 6778, 4568, 3131, 3716, 2907, 1202, 211, 14, 1 },
      { 10368, 6821, 4571, 3117, 3675, 2844, 1158, 200, 13, 1 },
      { 10496, 6862, 4573, 3102, 3634, 2783, 1116, 189, 12, 1 },
      { 10624, 6903, 4574, 3087, 3592, 2722, 1075, 179, 11, 1 },
      { 10752, 6942, 4575, 3071, 3551, 2662, 1035, 169, 10, 1 },
      { 10880, 6981, 4574, 3054, 3508, 2603, 997, 160, 10, 1 },
      { 11008, 7019, 4572, 3038, 3466, 2544, 960, 151, 9, 1 },
      { 11136, 7055, 4570, 3020, 3424, 2487, 924, 143, 8, 1 },
      { 11264, 7091, 4566, 3002, 3381, 2430, 890, 135, 8, 1 },
      { 11392, 7126, 4563, 2984, 3338, 2374, 856, 127, 7, 1 },
      { 11520, 7159, 4557, 2965, 3295, 2319, 824, 121, 7, 1 },
      { 11648, 7193, 4552, 2945, 3252, 2264, 793, 114, 6, 1 },
      { 11776, 7224, 4545, 2925, 3209, 2211, 763, 108, 6, 1 },
      { 11904, 7255, 4538, 2905, 3165, 2159, 734, 102, 5, 1 },
      { 12032, 7285, 4530, 2884, 3122, 2107, 706, 96, 5, 1 },
      { 12160, 7314, 4520, 2863, 3079, 2056, 679, 91, 5, 1 },
      { 12288, 7341, 4511, 2842, 3036, 2006, 653, 86, 4, 1 },
      { 12416, 7368, 4500, 2820, 2993, 1957, 628, 81, 4, 1 },
      { 12544, 7394, 4489, 2797, 2949, 1909, 604, 77, 4, 1 },
      { 12672, 7419, 4477, 2775, 2906, 1861, 581, 73, 3, 1 },
      { 12800, 7443, 4464, 2752, 2863, 1815, 558, 69, 3, 1 },
      { 12928, 7466, 4451, 2729, 2820, 1769, 536, 65, 3, 1 },
      { 13056, 7488, 4437, 2705, 2777, 1724, 516, 61, 3, 1 },
      { 13184, 7509, 4422, 2682, 2734, 1680, 495, 58, 3, 1 },
      { 13312, 7529, 4406, 2658, 2692, 1637, 476, 55, 2, 1 },
      { 13440, 7548, 4390, 2633, 2650, 1595, 457, 52, 2, 1 },
      { 13568, 7567, 4373, 2609, 2607, 1553, 439, 49, 2, 1 },
      { 13696, 7583, 4356, 2584, 2565, 1513, 422, 46, 2, 1 },
      { 13824, 7600, 4337, 2559, 2523, 1473, 405, 44, 2, 1 },
      { 13952, 7615, 4319, 2533, 2482, 1434, 389, 41, 2, 1 },
      { 14080, 7629, 4300, 2508, 2441, 1395, 373, 39, 2, 1 },
      { 14208, 7643, 4280, 2482, 2400, 1358, 358, 37, 1, 1 },
      { 14336, 7655, 4259, 2457, 2359, 1321, 344, 35, 1, 1 },
      { 14464, 7667, 4238, 2431, 2318, 1285, 330, 33, 1, 1 },
      { 14592, 7677, 4217, 2405, 2278, 1250, 316, 31, 1, 1 },
      { 14720, 7687, 4195, 2378, 2238, 1215, 304, 29, 1, 1 },
      { 14848, 7696, 4172, 2352, 2198, 1181, 291, 28, 1, 1 },
      { 14976, 7703, 4149, 2326, 2159, 1148, 279, 26, 1, 1 },
      { 15104, 7710, 4125, 2299, 2119, 1116, 268, 25, 1, 1 },
      { 15232, 7715, 4101, 2272, 2081, 1085, 257, 23, 1, 1 },
      { 15360, 7721, 4076, 2245, 2042, 1054, 246, 22, 1, 1 },
      { 15488, 7724, 4051, 2219, 2004, 1023, 236, 21, 1, 1 },
      { 15616, 7727, 4025, 2192, 1966, 994, 226, 20, 1, 1 },
      { 15744, 7729, 3999, 2164, 1929, 965, 217, 19, 1, 1 },
      { 15872, 7731, 3972, 2137, 1892, 937, 207, 18, 1, 1 },
      { 16000, 7731, 3945, 2110, 1855, 909, 199, 17, 1, 1 },
      { 16128, 7730, 3918, 2083, 1819, 882, 190, 16, 1, 1 },
      { 16256, 7728, 3890, 2056, 1783, 856, 182, 15, 1, 1 },
      { 16384, 7725, 3862, 2029, 1747, 831, 174, 14, 1, 1 },
      { 16512, 7721, 3833, 2002, 1712, 806, 167, 13, 1, 1 },
      { 16640, 7717, 3804, 1975, 1677, 781, 160, 12, 1, 1 },
      { 16768, 7712, 3775, 1947, 1642, 757, 153, 12, 1, 1 },
      { 16896, 7706, 3745, 1920, 1608, 734, 146, 11, 1, 1 },
      { 17024, 7699, 3714, 1893, 1575, 711, 140, 10, 1, 1 },
      { 17152, 7690, 3684, 1866, 1541, 689, 134, 10, 1, 1 },
      { 17280, 7681, 3653, 1839, 1508, 668, 128, 9, 1, 1 },
      { 17408, 7671, 3621, 1812, 1476, 647, 122, 9, 1, 1 },
      { 17536, 7660, 3590, 1785, 1444, 626, 117, 8, 1, 1 },
      { 17664, 7648, 3558, 1758, 1412, 606, 112, 8, 1, 1 },
      { 17792, 7635, 3526, 1731, 1381, 587, 107, 7, 1, 1 },
      { 17920, 7622, 3493, 1704, 1350, 568, 102, 7, 1, 1 },
      { 18048, 7607, 3461, 1678, 1319, 549, 98, 6, 1, 1 },
      { 18176, 7592, 3428, 1651, 1289, 531, 93, 6, 1, 1 },
      { 18304, 7575, 3394, 1625, 1259, 514, 89, 6, 1, 1 },
      { 18432, 7558, 3361, 1598, 1230, 497, 85, 5, 1, 1 },
      { 18560, 7540, 3327, 1572, 1201, 480, 81, 5, 1, 1 },
      { 18688, 7520, 3293, 1546, 1173, 464, 77, 5, 1, 1 },
      { 18816, 7500, 3258, 1520, 1145, 448, 74, 5, 1, 1 },
      { 18944, 7480, 3224, 1494, 1117, 433, 70, 4, 1, 1 },
      { 19072, 7458, 3189, 1468, 1090, 418, 67, 4, 1, 1 },
      { 19200, 7435, 3154, 1442, 1063, 404, 64, 4, 1, 1 },
      { 19328, 7410, 3119, 1417, 1037, 390, 61, 4, 1, 1 },
      { 19456, 7386, 3084, 1392, 1011, 376, 58, 3, 1, 1 },
      { 19584, 7361, 3048, 1366, 986, 363, 55, 3, 1, 1 },
      { 19712, 7335, 3012, 1341, 960, 350, 53, 3, 1, 1 },
      { 19840, 7307, 2977, 1316, 936, 337, 50, 3, 1, 1 },
      { 19968, 7279, 2941, 1291, 911, 325, 48, 3, 1, 1 },
      { 20096, 7251, 2905, 1267, 887, 313, 45, 2, 1, 1 },
      { 20224, 7220, 2868, 1243, 864, 302, 43, 2, 1, 1 },
      { 20352, 7189, 2832, 1218, 841, 291, 41, 2, 1, 1 },
      { 20480, 7158, 2795, 1194, 818, 280, 39, 2, 1, 1 },
      { 20608, 7124, 2759, 1170, 796, 270, 37, 2, 1, 1 },
      { 20736, 7091, 2722, 1147, 774, 259, 35, 2, 1, 1 },
      { 20864, 7056, 2685, 1123, 752, 250, 34, 2, 1, 1 },
      { 20992, 7021, 2648, 1100, 731, 240, 32, 2, 1, 1 },
      { 21120, 6985, 2612, 1077, 710, 231, 30, 1, 1, 1 },
      { 21248, 6948, 2574, 1054, 690, 222, 29, 1, 1, 1 },
      { 21376, 6911, 2537, 1031, 670, 213, 27, 1, 1, 1 },
      { 21504, 6872, 2500, 1008, 650, 205, 26, 1, 1, 1 },
      { 21632, 6831, 2463, 986, 631, 197, 25, 1, 1, 1 },
      { 21760, 6791, 2426, 964, 612, 189, 23, 1, 1, 1 },
      { 21888, 6749, 2389, 942, 594, 181, 22, 1, 1, 1 },
      { 22016, 6707, 2351, 921, 575, 174, 21, 1, 1, 1 },
      { 22144, 6663, 2314, 899, 558, 167, 20, 1, 1, 1 },
      { 22272, 6619, 2277, 878, 540, 160, 19, 1, 1, 1 },
      { 22400, 6574, 2240, 857, 523, 153, 18, 1, 1, 1 },
      { 22528, 6529, 2202, 836, 507, 146, 17, 1, 1, 1 },
      { 22656, 6482, 2165, 816, 490, 140, 16, 1, 1, 1 },
      { 22784, 6435, 2128, 795, 474, 134, 15, 1, 1, 1 },
      { 22912, 6386, 2091, 775, 459, 128, 14, 1, 1, 1 },
      { 23040, 6336, 2054, 756, 443, 123, 13, 1, 1, 1 },
      { 23168, 6286, 2017, 736, 428, 117, 13, 1, 1, 1 },
      { 23296, 6234, 1980, 717, 414, 112, 12, 1, 1, 1 },
      { 23424, 6183, 1943, 698, 399, 107, 11, 1, 1, 1 },
      { 23552, 6130, 1906, 679, 385, 102, 11, 1, 1, 1 },
      { 23680, 6077, 1869, 660, 372, 97, 10, 1, 1, 1 },
      { 23808, 6022, 1833, 642, 358, 93, 9, 1, 1, 1 },
      { 23936, 5966, 1796, 624, 345, 89, 9, 1, 1, 1 },
      { 24064, 5910, 1760, 606, 333, 84, 8, 1, 1, 1 },
      { 24192, 5853, 1724, 588, 320, 80, 8, 1, 1, 1 },
      { 24320, 5796, 1687, 571, 308, 76, 7, 1, 1, 1 },
      { 24448, 5735, 1651, 554, 297, 73, 7, 1, 1, 1 },
      { 24576, 5677, 1615, 537, 285, 69, 6, 1, 1, 1 },
      { 24704, 5615, 1579, 521, 274, 66, 6, 1, 1, 1 },
      { 24832, 5554, 1544, 504, 263, 62, 6, 1, 1, 1 },
      { 24960, 5492, 1508, 488, 253, 59, 5, 1, 1, 1 },
      { 25088, 5428, 1473, 473, 242, 56, 5, 1, 1, 1 },
      { 25216, 5364, 1438, 457, 232, 53, 5, 1, 1, 1 },
      { 25344, 5300, 1403, 442, 222, 50, 4, 1, 1, 1 },
      { 25472, 5233, 1368, 427, 213, 48, 4, 1, 1, 1 },
      { 25600, 5166, 1334, 412, 204, 45, 4, 1, 1, 1 },
      { 25728, 5098, 1299, 398, 195, 43, 4, 1, 1, 1 },
      { 25856, 5030, 1266, 384, 186, 40, 3, 1, 1, 1 },
      { 25984, 4960, 1232, 370, 178, 38, 3, 1, 1, 1 },
      { 26112, 4890, 1198, 356, 170, 36, 3, 1, 1, 1 },
      { 26240, 4819, 1164, 343, 162, 34, 3, 1, 1, 1 },
      { 26368, 4748, 1132, 329, 154, 32, 2, 1, 1, 1 },
      { 26496, 4675, 1098, 317, 147, 30, 2, 1, 1, 1 },
      { 26624, 4602, 1066, 304, 139, 28, 2, 1, 1, 1 },
      { 26752, 4527, 1034, 292, 132, 26, 2, 1, 1, 1 },
      { 26880, 4451, 1001, 280, 126, 25, 2, 1, 1, 1 },
      { 27008, 4375, 970, 268, 119, 23, 2, 1, 1, 1 },
      { 27136, 4299, 938, 256, 113, 21, 2, 1, 1, 1 },
      { 27264, 4221, 907, 245, 107, 20, 1, 1, 1, 1 },
      { 27392, 4142, 876, 234, 101, 19, 1, 1, 1, 1 },
      { 27520, 4063, 846, 223, 95, 17, 1, 1, 1, 1 },
      { 27648, 3982, 815, 213, 90, 16, 1, 1, 1, 1 },
      { 27776, 3900, 786, 202, 85, 15, 1, 1, 1, 1 },
      { 27904, 3818, 756, 192, 80, 14, 1, 1, 1, 1 },
      { 28032, 3734, 727, 183, 75, 13, 1, 1, 1, 1 },
      { 28160, 3651, 698, 173, 70, 12, 1, 1, 1, 1 },
      { 28288, 3566, 669, 164, 66, 11, 1, 1, 1, 1 },
      { 28416, 3481, 641, 155, 61, 10, 1, 1, 1, 1 },
      { 28544, 3393, 614, 147, 57, 9, 1, 1, 1, 1 },
      { 28672, 3306, 586, 138, 53, 9, 1, 1, 1, 1 },
      { 28800, 3217, 559, 130, 50, 8, 1, 1, 1, 1 },
      { 28928, 3128, 533, 122, 46, 7, 1, 1, 1, 1 },
      { 29056, 3037, 507, 114, 43, 7, 1, 1, 1, 1 },
      { 29184, 2947, 481, 107, 39, 6, 1, 1, 1, 1 },
      { 29312, 2855, 456, 100, 36, 5, 1, 1, 1, 1 },
      { 29440, 2762, 431, 93, 33, 5, 1, 1, 1, 1 },
      { 29568, 2668, 407, 86, 31, 4, 1, 1, 1, 1 },
      { 29696, 2573, 383, 80, 28, 4, 1, 1, 1, 1 },
      { 29824, 2478, 359, 74, 25, 4, 1, 1, 1, 1 },
      { 29952, 2381, 337, 68, 23, 3, 1, 1, 1, 1 },
      { 30080, 2284, 314, 62, 21, 3, 1, 1, 1, 1 },
      { 30208, 2185, 293, 57, 19, 2, 1, 1, 1, 1 },
      { 30336, 2086, 271, 52, 17, 2, 1, 1, 1, 1 },
      { 30464, 1986, 250, 47, 15, 2, 1, 1, 1, 1 },
      { 30592, 1885, 230, 42, 13, 2, 1, 1, 1, 1 },
      { 30720, 1782, 211, 38, 12, 1, 1, 1, 1, 1 },
      { 30848, 1679, 192, 34, 10, 1, 1, 1, 1, 1 },
      { 30976, 1575, 173, 30, 9, 1, 1, 1, 1, 1 },
      { 31104, 1469, 156, 26, 8, 1, 1, 1, 1, 1 },
      { 31232, 1364, 138, 23, 6, 1, 1, 1, 1, 1 },
      { 31360, 1257, 122, 19, 5, 1, 1, 1, 1, 1 },
      { 31488, 1149, 106, 16, 4, 1, 1, 1, 1, 1 },
      { 31616, 1038, 91, 14, 4, 1, 1, 1, 1, 1 },
      { 31744, 928, 77, 11, 3, 1, 1, 1, 1, 1 },
      { 31872, 816, 64, 9, 2, 1, 1, 1, 1, 1 },
      { 32000, 703, 51, 7, 2, 1, 1, 1, 1, 1 },
      { 32128, 589, 40, 5, 1, 1, 1, 1, 1, 1 },
      { 32256, 473, 29, 4, 1, 1, 1, 1, 1, 1 },
      { 32384, 357, 19, 2, 1, 1, 1, 1, 1, 1 },
      { 32512, 238, 11, 1, 1, 1, 1, 1, 1, 1 },
      { 32640, 117, 4, 1, 1, 1, 1, 1, 1, 1 },
    };
#endif  // CONFIG_EC_MULTISYMBOL

/* clang-format off */
#if CONFIG_ENTROPY
const av1_coeff_probs_model
default_qctx_coef_probs[QCTX_BINS][TX_SIZES][PLANE_TYPES] = {
    {  // Q_Index 0
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {182,  34, 137}, { 79,  39, 103}, { 10,  28,  51},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 45,  88, 147}, { 46,  80, 140}, { 25,  69, 119},
                        { 12,  57,  96}, {  4,  41,  65}, {  1,  20,  31},
                    },
                    {  // band 2
                        { 58, 124, 190}, { 39, 106, 178}, { 16,  86, 147},
                        {  7,  69, 114}, {  3,  50,  80}, {  1,  25,  42},
                    },
                    {  // band 3
                        { 90, 138, 215}, { 54, 116, 198}, { 18,  86, 155},
                        {  5,  62, 112}, {  1,  38,  68}, {  1,  17,  30},
                    },
                    {  // band 4
                        {126, 149, 231}, { 82, 114, 211}, { 21,  80, 157},
                        {  6,  56, 105}, {  1,  36,  64}, {  1,  17,  31},
                    },
                    {  // band 5
                        {171,  56, 236}, {140,  54, 219}, { 57,  45, 167},
                        { 26,  36, 113}, { 11,  29,  72}, {  3,  18,  39},
                    },
                },
                {  // Intra
                    {  // band 0
                        {153, 122, 186}, {106, 109, 171}, { 36,  84, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 27, 151, 201}, { 34, 131, 199}, { 23, 102, 161},
                        { 10,  80, 120}, {  4,  52,  78}, {  1,  24,  37},
                    },
                    {  // band 2
                        { 43, 158, 213}, { 35, 133, 203}, {  8,  92, 151},
                        {  2,  64, 106}, {  1,  36,  60}, {  1,  13,  24},
                    },
                    {  // band 3
                        { 68, 167, 223}, { 36, 135, 211}, {  9,  94, 157},
                        {  2,  67, 112}, {  1,  40,  68}, {  1,  17,  31},
                    },
                    {  // band 4
                        {131, 146, 237}, { 72, 119, 223}, { 17,  82, 164},
                        {  4,  55, 107}, {  1,  34,  63}, {  1,  16,  29},
                    },
                    {  // band 5
                        {184,  68, 244}, {153,  59, 232}, { 68,  51, 179},
                        { 31,  40, 123}, { 13,  29,  77}, {  4,  17,  37},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {203,  41, 203}, {127,  56, 174}, { 49,  56, 127},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {110, 121, 217}, {119, 113, 213}, { 64,  95, 185},
                        { 30,  72, 144}, {  8,  42,  76}, {  2,  17,  25},
                    },
                    {  // band 2
                        {127, 159, 229}, {115, 134, 223}, { 36, 100, 189},
                        { 11,  75, 142}, {  3,  48,  83}, {  1,  19,  33},
                    },
                    {  // band 3
                        {150, 172, 241}, { 90, 133, 231}, { 28, 102, 192},
                        {  7,  81, 147}, {  1,  53,  91}, {  1,  25,  42},
                    },
                    {  // band 4
                        {184, 144, 248}, {114, 117, 237}, { 37,  89, 192},
                        { 10,  63, 130}, {  4,  42,  76}, {  1,  19,  38},
                    },
                    {  // band 5
                        {207,  79, 250}, {179,  74, 241}, { 83,  67, 199},
                        { 38,  51, 142}, { 17,  37,  97}, { 10,  14,  55},
                    },
                },
                {  // Inter
                    {  // band 0
                        {220,  82, 232}, {150,  93, 214}, { 66,  95, 177},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 160, 227}, {136, 141, 227}, { 67, 114, 190},
                        { 40,  94, 148}, { 21,  70, 107}, { 10,  43,  63},
                    },
                    {  // band 2
                        {124, 173, 235}, {105, 147, 226}, { 27, 107, 184},
                        { 10,  80, 142}, {  3,  50,  86}, {  1,  16,  32},
                    },
                    {  // band 3
                        {149, 179, 243}, { 89, 147, 234}, { 29, 112, 193},
                        {  9,  94, 157}, {  1,  64, 111}, {  1,  25,  43},
                    },
                    {  // band 4
                        {187, 153, 248}, {127, 130, 241}, { 52,  99, 202},
                        { 20,  79, 152}, {  4,  50,  93}, {  1,  19,  32},
                    },
                    {  // band 5
                        {215,  82, 251}, {195,  80, 246}, { 93,  70, 204},
                        { 39,  54, 147}, { 14,  33,  88}, {  6,  14,  39},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {116,  43, 131}, { 39,  41,  94}, {  4,  28,  47},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 28, 101, 141}, { 27,  95, 140}, { 18,  80, 121},
                        { 10,  61,  95}, {  4,  39,  60}, {  1,  19,  26},
                    },
                    {  // band 2
                        { 29, 150, 183}, { 19, 127, 175}, {  8,  98, 147},
                        {  3,  76, 115}, {  1,  55,  84}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 26, 168, 202}, { 12, 138, 188}, {  2,  98, 149},
                        {  1,  69, 110}, {  1,  40,  65}, {  1,  17,  25},
                    },
                    {  // band 4
                        { 33, 188, 225}, { 12, 155, 207}, {  2, 101, 155},
                        {  1,  65, 106}, {  1,  36,  60}, {  1,  18,  26},
                    },
                    {  // band 5
                        { 79, 205, 242}, { 30, 168, 224}, {  5, 106, 164},
                        {  1,  68, 110}, {  1,  39,  65}, {  1,  18,  28},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 96,  80, 201}, { 51,  88, 168}, { 14,  78, 116},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  6, 167, 216}, { 32, 152, 211}, { 24, 121, 182},
                        { 13,  98, 149}, { 12,  76, 108}, {  8,  48,  62},
                    },
                    {  // band 2
                        { 17, 176, 225}, { 13, 147, 209}, {  3,  96, 155},
                        {  1,  65, 108}, {  2,  43,  63}, {  2,  23,  25},
                    },
                    {  // band 3
                        { 18, 183, 232}, { 10, 153, 214}, {  1,  96, 154},
                        {  1,  63, 105}, {  1,  39,  59}, {  1,  21,  24},
                    },
                    {  // band 4
                        { 23, 191, 239}, {  8, 159, 221}, {  1,  97, 158},
                        {  1,  61, 105}, {  1,  37,  60}, {  1,  20,  26},
                    },
                    {  // band 5
                        { 70, 201, 243}, { 29, 163, 228}, {  4, 102, 169},
                        {  1,  67, 114}, {  1,  39,  66}, {  1,  17,  29},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {181,  38, 192}, { 95,  47, 151}, { 29,  49, 102},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 72, 131, 202}, { 93, 120, 205}, { 50, 103, 179},
                        { 24,  79, 143}, { 11,  47,  78}, {  7,  19,  25},
                    },
                    {  // band 2
                        { 84, 176, 221}, { 56, 144, 214}, { 21, 108, 182},
                        {  8,  83, 139}, {  3,  55,  90}, {  2,  27,  41},
                    },
                    {  // band 3
                        { 84, 195, 234}, { 42, 156, 222}, { 10, 109, 180},
                        {  4,  77, 133}, {  1,  48,  80}, {  1,  23,  35},
                    },
                    {  // band 4
                        { 89, 210, 238}, { 35, 165, 221}, {  6, 106, 172},
                        {  2,  70, 123}, {  1,  44,  74}, {  1,  21,  30},
                    },
                    {  // band 5
                        {114, 221, 247}, { 49, 170, 234}, {  7, 113, 184},
                        {  2,  77, 132}, {  1,  48,  79}, {  1,  25,  33},
                    },
                },
                {  // Inter
                    {  // band 0
                        {192,  66, 237}, {113,  84, 211}, { 35,  84, 154},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 81, 180, 234}, {127, 165, 229}, { 58, 137, 204},
                        { 41, 114, 174}, { 44,  94, 136}, { 29,  66,  86},
                    },
                    {  // band 2
                        { 82, 193, 240}, { 39, 162, 223}, {  8, 113, 179},
                        {  3,  83, 136}, {  6,  62,  84}, {  5,  45,  45},
                    },
                    {  // band 3
                        { 78, 203, 242}, { 31, 170, 227}, {  4, 115, 181},
                        {  1,  82, 135}, {  2,  59,  82}, {  1,  45,  47},
                    },
                    {  // band 4
                        { 76, 210, 239}, { 25, 170, 213}, {  2,  99, 152},
                        {  1,  69, 115}, {  1,  49,  80}, {  1,  47,  57},
                    },
                    {  // band 5
                        {103, 217, 250}, { 42, 180, 237}, {  3, 124, 191},
                        {  1,  90, 150}, {  1,  69, 116}, {  1,  52,  46},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 58,  38,  99}, {  9,  26,  51}, {  1,  14,  22},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 14,  78, 109}, { 16,  73, 105}, { 11,  62,  92},
                        {  6,  47,  72}, {  2,  29,  45}, {  1,  12,  18},
                    },
                    {  // band 2
                        { 17, 131, 148}, { 11, 112, 140}, {  5,  87, 118},
                        {  2,  63,  90}, {  1,  42,  63}, {  1,  19,  31},
                    },
                    {  // band 3
                        { 12, 151, 168}, {  6, 116, 152}, {  1,  76, 115},
                        {  1,  50,  81}, {  1,  32,  52}, {  1,  14,  23},
                    },
                    {  // band 4
                        { 10, 174, 191}, {  3, 130, 172}, {  1,  80, 126},
                        {  1,  53,  88}, {  1,  32,  55}, {  1,  14,  24},
                    },
                    {  // band 5
                        { 19, 219, 237}, {  3, 168, 211}, {  1,  90, 142},
                        {  1,  53,  91}, {  1,  29,  51}, {  1,  12,  21},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 21,  46, 184}, { 10,  53, 130}, {  2,  49,  78},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  3, 169, 198}, { 37, 165, 196}, { 26, 134, 176},
                        { 11, 108, 149}, {  5,  81, 112}, {  3,  47,  64},
                    },
                    {  // band 2
                        { 11, 183, 215}, {  8, 142, 192}, {  2,  91, 141},
                        {  1,  62, 100}, {  1,  38,  62}, {  1,  17,  28},
                    },
                    {  // band 3
                        { 12, 190, 223}, {  6, 149, 199}, {  1,  88, 139},
                        {  1,  56,  93}, {  1,  31,  54}, {  1,  13,  21},
                    },
                    {  // band 4
                        { 11, 197, 230}, {  3, 154, 204}, {  1,  83, 134},
                        {  1,  50,  86}, {  1,  28,  49}, {  1,  12,  21},
                    },
                    {  // band 5
                        { 17, 211, 240}, {  2, 167, 217}, {  1,  88, 143},
                        {  1,  53,  91}, {  1,  30,  53}, {  1,  14,  24},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {151,  30, 151}, { 50,  36, 105}, {  8,  34,  66},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 39, 111, 160}, { 62, 111, 165}, { 37,  99, 147},
                        { 15,  77, 118}, {  3,  47,  73}, {  1,  17,  27},
                    },
                    {  // band 2
                        { 48, 170, 190}, { 32, 135, 180}, { 11, 100, 149},
                        {  4,  76, 116}, {  1,  51,  80}, {  1,  22,  36},
                    },
                    {  // band 3
                        { 39, 191, 208}, { 18, 141, 191}, {  3,  96, 150},
                        {  1,  66, 110}, {  1,  41,  69}, {  1,  17,  28},
                    },
                    {  // band 4
                        { 32, 209, 219}, {  8, 152, 201}, {  1,  96, 153},
                        {  1,  63, 106}, {  1,  38,  66}, {  1,  17,  29},
                    },
                    {  // band 5
                        { 33, 230, 237}, {  5, 173, 214}, {  1, 100, 155},
                        {  1,  62, 105}, {  1,  38,  66}, {  1,  18,  32},
                    },
                },
                {  // Inter
                    {  // band 0
                        {149,  38, 231}, { 59,  51, 186}, { 12,  54, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 53, 179, 226}, {126, 176, 223}, { 58, 147, 202},
                        { 28, 118, 174}, { 15,  94, 138}, { 14,  63,  87},
                    },
                    {  // band 2
                        { 58, 196, 232}, { 26, 158, 213}, {  5, 106, 166},
                        {  1,  75, 124}, {  1,  46,  79}, {  1,  23,  39},
                    },
                    {  // band 3
                        { 46, 203, 235}, { 17, 162, 213}, {  2, 104, 165},
                        {  1,  72, 120}, {  1,  44,  74}, {  1,  22,  33},
                    },
                    {  // band 4
                        { 37, 213, 238}, {  8, 167, 216}, {  1, 104, 168},
                        {  1,  68, 119}, {  1,  40,  67}, {  1,  17,  29},
                    },
                    {  // band 5
                        { 30, 228, 239}, {  4, 181, 213}, {  1, 103, 153},
                        {  1,  65, 110}, {  1,  43,  79}, {  1,  27,  56},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 76,  25,  53}, {  9,  18,  32}, {  1,  12,  18},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 29,  55,  91}, { 19,  58,  95}, { 15,  57,  89},
                        { 12,  49,  77}, {  3,  29,  44}, {  1,   8,  12},
                    },
                    {  // band 2
                        { 32, 160, 148}, { 33, 143, 146}, { 19, 122, 132},
                        {  6,  90, 102}, {  1,  58,  70}, {  1,  17,  24},
                    },
                    {  // band 3
                        { 16, 181, 181}, {  6, 142, 165}, {  1,  90, 120},
                        {  1,  50,  71}, {  1,  25,  38}, {  1,   9,  14},
                    },
                    {  // band 4
                        { 13, 203, 203}, {  3, 154, 176}, {  1,  80, 108},
                        {  1,  41,  61}, {  1,  24,  37}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 234, 240}, {  1, 178, 204}, {  1,  80, 119},
                        {  1,  45,  71}, {  1,  26,  42}, {  1,  12,  19},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 78,  20, 135}, { 25,  18, 101}, {  5,  19,  57},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  7, 144, 183}, {117, 151, 195}, {109, 151, 187},
                        { 39, 130, 168}, { 11, 100, 125}, {  4,  59,  64},
                    },
                    {  // band 2
                        { 20, 184, 212}, { 12, 148, 191}, {  2,  98, 141},
                        {  1,  65, 100}, {  1,  39,  61}, {  1,  14,  22},
                    },
                    {  // band 3
                        { 15, 194, 222}, {  6, 153, 198}, {  1,  92, 138},
                        {  1,  58,  91}, {  1,  32,  52}, {  1,  12,  18},
                    },
                    {  // band 4
                        { 14, 206, 232}, {  3, 162, 206}, {  1,  89, 134},
                        {  1,  52,  83}, {  1,  28,  46}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 225, 241}, {  1, 175, 210}, {  1,  81, 125},
                        {  1,  48,  78}, {  1,  28,  46}, {  1,  13,  21},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {124,  23,  93}, { 31,  24,  63}, {  6,  24,  46},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 23,  86, 126}, { 45,  90, 145}, { 31,  91, 133},
                        { 19,  80, 114}, {  7,  53,  72}, {  1,  20,  27},
                    },
                    {  // band 2
                        { 51, 186, 189}, { 48, 159, 182}, { 33, 128, 156},
                        { 15,  92, 124}, {  2,  62,  83}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 36, 198, 211}, { 15, 156, 187}, {  3,  97, 137},
                        {  1,  61,  93}, {  1,  35,  57}, {  1,  15,  23},
                    },
                    {  // band 4
                        { 34, 219, 223}, {  9, 162, 193}, {  1,  91, 136},
                        {  1,  58,  92}, {  1,  35,  54}, {  1,  14,  23},
                    },
                    {  // band 5
                        { 19, 243, 243}, {  3, 191, 208}, {  1,  91, 137},
                        {  1,  56,  90}, {  1,  34,  55}, {  1,  16,  24},
                    },
                },
                {  // Inter
                    {  // band 0
                        {119,  20, 197}, { 19,  29, 156}, {  3,  30, 107},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 24, 192, 226}, {161, 193, 227}, { 97, 185, 222},
                        { 31, 158, 204}, { 16, 122, 165}, { 17,  84, 112},
                    },
                    {  // band 2
                        { 26, 202, 229}, { 11, 165, 210}, {  2, 103, 152},
                        {  1,  68, 104}, {  1,  42,  70}, {  1,  16,  36},
                    },
                    {  // band 3
                        { 24, 209, 237}, {  6, 169, 214}, {  1, 102, 154},
                        {  1,  65, 107}, {  1,  45,  68}, {  1,  17,  24},
                    },
                    {  // band 4
                        { 19, 219, 243}, {  4, 183, 226}, {  1, 115, 172},
                        {  1,  73, 119}, {  1,  43,  77}, {  1,  15,  37},
                    },
                    {  // band 5
                        { 11, 237, 241}, {  2, 190, 216}, {  1, 108, 146},
                        {  1,  59,  94}, {  1,  40,  67}, {  1,  30,  53},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 76,  25,  53}, {  9,  18,  32}, {  1,  12,  18},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 29,  55,  91}, { 19,  58,  95}, { 15,  57,  89},
                        { 12,  49,  77}, {  3,  29,  44}, {  1,   8,  12},
                    },
                    {  // band 2
                        { 32, 160, 148}, { 33, 143, 146}, { 19, 122, 132},
                        {  6,  90, 102}, {  1,  58,  70}, {  1,  17,  24},
                    },
                    {  // band 3
                        { 16, 181, 181}, {  6, 142, 165}, {  1,  90, 120},
                        {  1,  50,  71}, {  1,  25,  38}, {  1,   9,  14},
                    },
                    {  // band 4
                        { 13, 203, 203}, {  3, 154, 176}, {  1,  80, 108},
                        {  1,  41,  61}, {  1,  24,  37}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 234, 240}, {  1, 178, 204}, {  1,  80, 119},
                        {  1,  45,  71}, {  1,  26,  42}, {  1,  12,  19},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 78,  20, 135}, { 25,  18, 101}, {  5,  19,  57},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {  7, 144, 183}, {117, 151, 195}, {109, 151, 187},
                        { 39, 130, 168}, { 11, 100, 125}, {  4,  59,  64},
                    },
                    {  // band 2
                        { 20, 184, 212}, { 12, 148, 191}, {  2,  98, 141},
                        {  1,  65, 100}, {  1,  39,  61}, {  1,  14,  22},
                    },
                    {  // band 3
                        { 15, 194, 222}, {  6, 153, 198}, {  1,  92, 138},
                        {  1,  58,  91}, {  1,  32,  52}, {  1,  12,  18},
                    },
                    {  // band 4
                        { 14, 206, 232}, {  3, 162, 206}, {  1,  89, 134},
                        {  1,  52,  83}, {  1,  28,  46}, {  1,  11,  17},
                    },
                    {  // band 5
                        {  6, 225, 241}, {  1, 175, 210}, {  1,  81, 125},
                        {  1,  48,  78}, {  1,  28,  46}, {  1,  13,  21},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {124,  23,  93}, { 31,  24,  63}, {  6,  24,  46},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 23,  86, 126}, { 45,  90, 145}, { 31,  91, 133},
                        { 19,  80, 114}, {  7,  53,  72}, {  1,  20,  27},
                    },
                    {  // band 2
                        { 51, 186, 189}, { 48, 159, 182}, { 33, 128, 156},
                        { 15,  92, 124}, {  2,  62,  83}, {  1,  29,  43},
                    },
                    {  // band 3
                        { 36, 198, 211}, { 15, 156, 187}, {  3,  97, 137},
                        {  1,  61,  93}, {  1,  35,  57}, {  1,  15,  23},
                    },
                    {  // band 4
                        { 34, 219, 223}, {  9, 162, 193}, {  1,  91, 136},
                        {  1,  58,  92}, {  1,  35,  54}, {  1,  14,  23},
                    },
                    {  // band 5
                        { 19, 243, 243}, {  3, 191, 208}, {  1,  91, 137},
                        {  1,  56,  90}, {  1,  34,  55}, {  1,  16,  24},
                    },
                },
                {  // Inter
                    {  // band 0
                        {119,  20, 197}, { 19,  29, 156}, {  3,  30, 107},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 24, 192, 226}, {161, 193, 227}, { 97, 185, 222},
                        { 31, 158, 204}, { 16, 122, 165}, { 17,  84, 112},
                    },
                    {  // band 2
                        { 26, 202, 229}, { 11, 165, 210}, {  2, 103, 152},
                        {  1,  68, 104}, {  1,  42,  70}, {  1,  16,  36},
                    },
                    {  // band 3
                        { 24, 209, 237}, {  6, 169, 214}, {  1, 102, 154},
                        {  1,  65, 107}, {  1,  45,  68}, {  1,  17,  24},
                    },
                    {  // band 4
                        { 19, 219, 243}, {  4, 183, 226}, {  1, 115, 172},
                        {  1,  73, 119}, {  1,  43,  77}, {  1,  15,  37},
                    },
                    {  // band 5
                        { 11, 237, 241}, {  2, 190, 216}, {  1, 108, 146},
                        {  1,  59,  94}, {  1,  40,  67}, {  1,  30,  53},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 1
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {174,  30, 159}, { 76,  38, 115}, { 15,  33,  65},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 60,  80, 153}, { 72,  75, 147}, { 36,  68, 129},
                        { 15,  59, 104}, {  4,  45,  74}, {  1,  28,  45},
                    },
                    {  // band 2
                        { 70, 122, 186}, { 55, 104, 175}, { 21,  83, 144},
                        {  8,  67, 112}, {  2,  51,  82}, {  1,  34,  57},
                    },
                    {  // band 3
                        { 97, 144, 207}, { 52, 109, 195}, { 16,  77, 153},
                        {  4,  58, 113}, {  1,  43,  77}, {  1,  27,  48},
                    },
                    {  // band 4
                        {128, 148, 229}, { 76, 104, 210}, { 18,  77, 159},
                        {  4,  65, 110}, {  1,  52,  82}, {  1,  31,  55},
                    },
                    {  // band 5
                        {165,  51, 238}, {128,  50, 230}, { 57,  49, 185},
                        { 28,  47, 130}, { 12,  44,  96}, {  3,  36,  60},
                    },
                },
                {  // Intra
                    {  // band 0
                        {169, 103, 203}, {117,  96, 176}, { 56,  81, 137},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 31, 150, 224}, { 49, 128, 212}, { 19,  92, 165},
                        {  6,  67, 116}, {  2,  43,  71}, {  1,  21,  36},
                    },
                    {  // band 2
                        { 58, 156, 230}, { 47, 130, 215}, {  7,  87, 158},
                        {  2,  63, 114}, {  1,  39,  71}, {  1,  18,  36},
                    },
                    {  // band 3
                        { 85, 176, 240}, { 43, 138, 226}, {  8,  93, 172},
                        {  2,  70, 127}, {  1,  46,  81}, {  1,  26,  47},
                    },
                    {  // band 4
                        {155, 144, 248}, { 93, 116, 235}, { 21,  83, 180},
                        {  4,  59, 119}, {  1,  43,  80}, {  1,  25,  50},
                    },
                    {  // band 5
                        {203,  61, 250}, {171,  57, 243}, { 71,  57, 199},
                        { 31,  49, 144}, { 13,  42,  96}, {  7,  30,  52},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {204,  44, 204}, {137,  57, 184}, { 72,  62, 152},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {145, 117, 236}, {151, 112, 231}, { 87,  95, 208},
                        { 31,  77, 165}, {  5,  49,  98}, {  1,  24,  39},
                    },
                    {  // band 2
                        {146, 152, 241}, {140, 132, 236}, { 41, 103, 209},
                        { 10,  86, 165}, {  2,  55, 106}, {  1,  25,  58},
                    },
                    {  // band 3
                        {154, 181, 249}, { 84, 143, 240}, { 23, 114, 210},
                        {  6, 102, 182}, {  2,  71, 137}, {  1,  35,  90},
                    },
                    {  // band 4
                        {184, 150, 251}, {115, 130, 244}, { 34, 105, 215},
                        { 15,  89, 173}, {  1,  51, 141}, {128, 128, 128},
                    },
                    {  // band 5
                        {211,  71, 253}, {193,  78, 249}, {106,  91, 232},
                        { 61,  87, 198}, { 85, 153, 254}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {232, 104, 242}, {165, 114, 227}, { 96, 120, 206},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {137, 178, 250}, {146, 153, 245}, { 74, 108, 205},
                        { 41,  81, 149}, { 24,  55, 104}, { 13,  36,  68},
                    },
                    {  // band 2
                        {147, 185, 252}, {127, 161, 246}, { 30, 104, 208},
                        { 11,  74, 154}, {  6,  54, 100}, {  2,  29,  63},
                    },
                    {  // band 3
                        {163, 191, 254}, {101, 161, 249}, { 22, 114, 215},
                        {  6,  89, 173}, {  1,  65, 120}, {  1,   1, 170},
                    },
                    {  // band 4
                        {197, 160, 254}, {142, 141, 251}, { 39, 102, 218},
                        { 10,  76, 158}, {  1,  56, 122}, {128, 128, 128},
                    },
                    {  // band 5
                        {224,  76, 254}, {215,  84, 253}, {107,  85, 232},
                        { 43,  71, 177}, {  1,   1, 254}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 68,  37, 120}, { 21,  34,  82}, {  5,  26,  49},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41,  89, 138}, { 56,  83, 132}, { 31,  73, 115},
                        { 16,  62,  92}, {  5,  45,  62}, {  1,  24,  32},
                    },
                    {  // band 2
                        { 48, 139, 165}, { 30, 114, 160}, { 13,  92, 132},
                        {  6,  72, 103}, {  3,  49,  72}, {  1,  26,  41},
                    },
                    {  // band 3
                        { 44, 162, 191}, { 20, 127, 175}, {  5,  90, 137},
                        {  1,  62, 100}, {  1,  38,  63}, {  1,  20,  32},
                    },
                    {  // band 4
                        { 51, 184, 213}, { 16, 137, 193}, {  2,  89, 143},
                        {  1,  60, 102}, {  1,  39,  66}, {  1,  23,  37},
                    },
                    {  // band 5
                        { 76, 200, 235}, { 27, 150, 216}, {  3,  99, 164},
                        {  1,  70, 119}, {  1,  45,  77}, {  1,  22,  38},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 81, 112, 199}, { 49, 101, 164}, { 19,  80, 119},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 12, 181, 217}, { 48, 151, 212}, { 38, 118, 180},
                        { 22,  95, 140}, { 11,  67,  92}, { 13,  46,  44},
                    },
                    {  // band 2
                        { 29, 188, 226}, { 19, 147, 210}, {  5,  95, 154},
                        {  4,  68, 106}, {  3,  44,  60}, {  1,  24,  27},
                    },
                    {  // band 3
                        { 30, 195, 234}, { 15, 153, 216}, {  3,  95, 156},
                        {  2,  66, 108}, {  2,  44,  62}, {  1,  24,  29},
                    },
                    {  // band 4
                        { 36, 203, 243}, { 12, 162, 225}, {  2,  98, 163},
                        {  2,  67, 113}, {  2,  45,  68}, {  1,  24,  34},
                    },
                    {  // band 5
                        { 86, 207, 248}, { 35, 165, 236}, {  3, 107, 180},
                        {  1,  73, 128}, {  1,  45,  78}, {  1,  20,  34},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {188,  37, 205}, {118,  51, 172}, { 56,  57, 135},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {116, 135, 225}, {144, 123, 221}, { 72, 103, 197},
                        { 35,  77, 153}, { 15,  47,  82}, {  6,  25,  34},
                    },
                    {  // band 2
                        {128, 171, 233}, { 82, 142, 226}, { 31, 106, 191},
                        { 16,  82, 146}, {  9,  59,  98}, {  4,  33,  54},
                    },
                    {  // band 3
                        {126, 197, 241}, { 66, 155, 230}, { 18, 108, 190},
                        {  7,  82, 148}, {  3,  58,  98}, {  1,  25,  50},
                    },
                    {  // band 4
                        {117, 207, 244}, { 44, 163, 233}, {  9, 112, 191},
                        {  5,  84, 148}, {  3,  61,  87}, {  1,  28,  38},
                    },
                    {  // band 5
                        {112, 214, 249}, { 39, 174, 240}, {  6, 125, 205},
                        {  4,  96, 163}, {  5,  66, 100}, {  1, 128, 254},
                    },
                },
                {  // Inter
                    {  // band 0
                        {227,  70, 234}, {145,  91, 213}, { 61, 100, 173},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {108, 198, 243}, {171, 172, 240}, {118, 130, 210},
                        {104, 107, 165}, { 64,  85, 114}, { 55,  64,  60},
                    },
                    {  // band 2
                        {110, 208, 247}, { 64, 175, 237}, { 24, 112, 187},
                        { 24,  81, 133}, { 24,  63,  83}, { 21,  47,  53},
                    },
                    {  // band 3
                        { 91, 218, 249}, { 46, 188, 238}, {  8, 113, 184},
                        {  5,  83, 137}, {  6,  62,  95}, { 17,  44,  94},
                    },
                    {  // band 4
                        { 84, 216, 248}, { 30, 187, 237}, {  2, 117, 188},
                        {  1,  88, 141}, {  3,  63,  98}, {  1,   1,   1},
                    },
                    {  // band 5
                        {116, 218, 252}, { 47, 186, 242}, {  2, 132, 204},
                        {  1, 106, 175}, {  1,  88, 104}, {  1, 254, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 35,  41, 129}, { 12,  30,  70}, {  2,  19,  32},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 30,  77, 116}, { 39,  70, 110}, { 20,  58,  96},
                        {  8,  47,  77}, {  2,  33,  52}, {  1,  17,  26},
                    },
                    {  // band 2
                        { 31, 123, 146}, { 18, 103, 140}, {  7,  81, 119},
                        {  2,  62,  95}, {  1,  44,  70}, {  1,  26,  42},
                    },
                    {  // band 3
                        { 21, 149, 170}, {  9, 114, 158}, {  2,  80, 126},
                        {  1,  57,  94}, {  1,  36,  61}, {  1,  18,  31},
                    },
                    {  // band 4
                        { 20, 178, 199}, {  6, 134, 183}, {  1,  87, 139},
                        {  1,  60, 100}, {  1,  37,  64}, {  1,  18,  31},
                    },
                    {  // band 5
                        { 36, 218, 233}, {  6, 160, 207}, {  1,  92, 147},
                        {  1,  59, 101}, {  1,  35,  62}, {  1,  18,  31},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 17,  62, 211}, { 14,  62, 153}, {  5,  50,  84},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 11, 180, 205}, { 87, 160, 205}, { 53, 128, 184},
                        { 27, 106, 156}, { 13,  79, 115}, {  6,  46,  67},
                    },
                    {  // band 2
                        { 32, 194, 220}, { 20, 145, 202}, {  4,  96, 152},
                        {  1,  67, 111}, {  1,  42,  70}, {  1,  21,  37},
                    },
                    {  // band 3
                        { 30, 204, 228}, { 14, 152, 207}, {  1,  92, 149},
                        {  1,  61, 103}, {  1,  34,  59}, {  1,  16,  28},
                    },
                    {  // band 4
                        { 27, 213, 235}, {  7, 159, 210}, {  1,  88, 143},
                        {  1,  55,  94}, {  1,  31,  53}, {  1,  16,  27},
                    },
                    {  // band 5
                        { 28, 223, 243}, {  4, 173, 217}, {  1,  91, 146},
                        {  1,  58,  98}, {  1,  35,  60}, {  1,  19,  33},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {172,  37, 202}, { 83,  51, 156}, { 24,  53, 110},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 134, 206}, {110, 124, 200}, { 47, 106, 180},
                        { 15,  82, 145}, {  3,  48,  83}, {  1,  19,  32},
                    },
                    {  // band 2
                        { 80, 176, 220}, { 49, 145, 212}, { 17, 112, 180},
                        {  7,  84, 140}, {  1,  53,  89}, {  1,  27,  43},
                    },
                    {  // band 3
                        { 74, 201, 232}, { 38, 158, 221}, {  8, 112, 179},
                        {  2,  79, 132}, {  1,  47,  82}, {  1,  26,  42},
                    },
                    {  // band 4
                        { 73, 215, 239}, { 28, 169, 227}, {  3, 112, 176},
                        {  1,  74, 126}, {  1,  48,  79}, {  1,  27,  44},
                    },
                    {  // band 5
                        { 71, 233, 244}, { 18, 180, 230}, {  1, 114, 180},
                        {  1,  80, 134}, {  1,  51,  85}, {  1,  26,  36},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  34, 244}, {126,  57, 212}, { 46,  67, 151},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {120, 202, 245}, {198, 173, 241}, {119, 146, 224},
                        { 76, 126, 195}, { 44, 102, 159}, { 40,  76, 115},
                    },
                    {  // band 2
                        {120, 215, 248}, { 69, 171, 237}, { 23, 119, 194},
                        { 10,  86, 147}, {  2,  56,  94}, {  1,  25,  44},
                    },
                    {  // band 3
                        {102, 226, 250}, { 53, 183, 239}, {  9, 118, 188},
                        {  2,  78, 131}, {  1,  48,  89}, {  1,  17,  36},
                    },
                    {  // band 4
                        { 86, 235, 252}, { 34, 194, 240}, {  2, 109, 173},
                        {  1,  68, 118}, {  1,  44,  79}, {  1,   1,  38},
                    },
                    {  // band 5
                        { 59, 236, 243}, { 11, 189, 228}, {  1, 112, 187},
                        {  1,  88, 145}, {  1,  55,  92}, {  1,   1, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 41,  40, 104}, { 12,  31,  64}, {  2,  16,  28},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 65,  58, 132}, { 50,  61, 130}, { 40,  57, 116},
                        { 22,  46,  87}, {  2,  28,  44}, {  1,  11,  17},
                    },
                    {  // band 2
                        { 55, 139, 135}, { 46, 122, 132}, { 21,  89, 110},
                        {  6,  60,  78}, {  1,  38,  54}, {  1,  17,  27},
                    },
                    {  // band 3
                        { 29, 167, 161}, { 10, 120, 141}, {  1,  69,  98},
                        {  1,  42,  66}, {  1,  28,  44}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 19, 191, 180}, {  4, 125, 154}, {  1,  70, 107},
                        {  1,  48,  77}, {  1,  33,  53}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 16, 238, 231}, {  2, 163, 198}, {  1,  85, 134},
                        {  1,  54,  90}, {  1,  34,  57}, {  1,  17,  29},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 70,  15, 216}, { 40,  18, 164}, { 14,  17,  83},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 25, 150, 200}, {185, 154, 211}, {123, 137, 199},
                        { 67, 119, 177}, { 31,  96, 137}, { 18,  63,  86},
                    },
                    {  // band 2
                        { 57, 187, 223}, { 35, 148, 207}, {  7, 104, 159},
                        {  2,  72, 113}, {  1,  44,  71}, {  1,  20,  34},
                    },
                    {  // band 3
                        { 44, 203, 233}, { 18, 157, 212}, {  1,  98, 150},
                        {  1,  61, 102}, {  1,  38,  62}, {  1,  19,  31},
                    },
                    {  // band 4
                        { 41, 215, 238}, { 11, 166, 215}, {  1,  94, 146},
                        {  1,  60, 101}, {  1,  37,  63}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 19, 236, 246}, {  3, 188, 223}, {  1,  95, 146},
                        {  1,  58,  95}, {  1,  34,  56}, {  1,  17,  27},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {146,  27, 156}, { 49,  32, 116}, { 10,  39,  77},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 47, 101, 172}, { 93, 100, 178}, { 58,  91, 165},
                        { 26,  75, 134}, {  4,  49,  82}, {  2,  22,  33},
                    },
                    {  // band 2
                        { 60, 158, 196}, { 44, 135, 186}, { 25, 106, 157},
                        {  8,  81, 124}, {  2,  56,  86}, {  1,  28,  45},
                    },
                    {  // band 3
                        { 44, 169, 212}, { 15, 138, 196}, {  2, 100, 157},
                        {  1,  74, 119}, {  1,  49,  76}, {  1,  20,  34},
                    },
                    {  // band 4
                        { 38, 199, 231}, { 11, 158, 214}, {  1, 111, 167},
                        {  1,  76, 122}, {  1,  44,  76}, {  1,  17,  39},
                    },
                    {  // band 5
                        { 40, 236, 246}, { 10, 187, 230}, {  1, 115, 175},
                        {  1,  74, 122}, {  1,  42,  71}, {  1,  14,  59},
                    },
                },
                {  // Inter
                    {  // band 0
                        {161,  26, 237}, { 65,  46, 209}, { 21,  46, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 87, 229, 245}, {206, 214, 244}, {148, 186, 236},
                        { 89, 165, 221}, { 41, 132, 186}, { 37,  93, 141},
                    },
                    {  // band 2
                        { 93, 231, 246}, { 47, 181, 231}, {  8, 117, 188},
                        {  2,  84, 138}, {  1,  43,  87}, {  1,  27,  41},
                    },
                    {  // band 3
                        { 80, 239, 250}, { 28, 190, 236}, {  1, 119, 183},
                        {  1,  84, 135}, {  1,  81,  69}, {  1, 102,   1},
                    },
                    {  // band 4
                        { 67, 245, 252}, { 22, 206, 242}, {  1, 130, 195},
                        {  1,  77, 136}, {  1,  35,  88}, {128, 128, 128},
                    },
                    {  // band 5
                        { 43, 250, 228}, { 31, 185, 204}, {  6, 101, 183},
                        {  1,  92, 151}, {  1,  84, 137}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 41,  40, 104}, { 12,  31,  64}, {  2,  16,  28},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 65,  58, 132}, { 50,  61, 130}, { 40,  57, 116},
                        { 22,  46,  87}, {  2,  28,  44}, {  1,  11,  17},
                    },
                    {  // band 2
                        { 55, 139, 135}, { 46, 122, 132}, { 21,  89, 110},
                        {  6,  60,  78}, {  1,  38,  54}, {  1,  17,  27},
                    },
                    {  // band 3
                        { 29, 167, 161}, { 10, 120, 141}, {  1,  69,  98},
                        {  1,  42,  66}, {  1,  28,  44}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 19, 191, 180}, {  4, 125, 154}, {  1,  70, 107},
                        {  1,  48,  77}, {  1,  33,  53}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 16, 238, 231}, {  2, 163, 198}, {  1,  85, 134},
                        {  1,  54,  90}, {  1,  34,  57}, {  1,  17,  29},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 70,  15, 216}, { 40,  18, 164}, { 14,  17,  83},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 25, 150, 200}, {185, 154, 211}, {123, 137, 199},
                        { 67, 119, 177}, { 31,  96, 137}, { 18,  63,  86},
                    },
                    {  // band 2
                        { 57, 187, 223}, { 35, 148, 207}, {  7, 104, 159},
                        {  2,  72, 113}, {  1,  44,  71}, {  1,  20,  34},
                    },
                    {  // band 3
                        { 44, 203, 233}, { 18, 157, 212}, {  1,  98, 150},
                        {  1,  61, 102}, {  1,  38,  62}, {  1,  19,  31},
                    },
                    {  // band 4
                        { 41, 215, 238}, { 11, 166, 215}, {  1,  94, 146},
                        {  1,  60, 101}, {  1,  37,  63}, {  1,  17,  28},
                    },
                    {  // band 5
                        { 19, 236, 246}, {  3, 188, 223}, {  1,  95, 146},
                        {  1,  58,  95}, {  1,  34,  56}, {  1,  17,  27},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {146,  27, 156}, { 49,  32, 116}, { 10,  39,  77},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 47, 101, 172}, { 93, 100, 178}, { 58,  91, 165},
                        { 26,  75, 134}, {  4,  49,  82}, {  2,  22,  33},
                    },
                    {  // band 2
                        { 60, 158, 196}, { 44, 135, 186}, { 25, 106, 157},
                        {  8,  81, 124}, {  2,  56,  86}, {  1,  28,  45},
                    },
                    {  // band 3
                        { 44, 169, 212}, { 15, 138, 196}, {  2, 100, 157},
                        {  1,  74, 119}, {  1,  49,  76}, {  1,  20,  34},
                    },
                    {  // band 4
                        { 38, 199, 231}, { 11, 158, 214}, {  1, 111, 167},
                        {  1,  76, 122}, {  1,  44,  76}, {  1,  17,  39},
                    },
                    {  // band 5
                        { 40, 236, 246}, { 10, 187, 230}, {  1, 115, 175},
                        {  1,  74, 122}, {  1,  42,  71}, {  1,  14,  59},
                    },
                },
                {  // Inter
                    {  // band 0
                        {161,  26, 237}, { 65,  46, 209}, { 21,  46, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 87, 229, 245}, {206, 214, 244}, {148, 186, 236},
                        { 89, 165, 221}, { 41, 132, 186}, { 37,  93, 141},
                    },
                    {  // band 2
                        { 93, 231, 246}, { 47, 181, 231}, {  8, 117, 188},
                        {  2,  84, 138}, {  1,  43,  87}, {  1,  27,  41},
                    },
                    {  // band 3
                        { 80, 239, 250}, { 28, 190, 236}, {  1, 119, 183},
                        {  1,  84, 135}, {  1,  81,  69}, {  1, 102,   1},
                    },
                    {  // band 4
                        { 67, 245, 252}, { 22, 206, 242}, {  1, 130, 195},
                        {  1,  77, 136}, {  1,  35,  88}, {128, 128, 128},
                    },
                    {  // band 5
                        { 43, 250, 228}, { 31, 185, 204}, {  6, 101, 183},
                        {  1,  92, 151}, {  1,  84, 137}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 2
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {181,  22, 175}, { 96,  37, 147}, { 35,  41, 105},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80,  95, 197}, {111,  92, 193}, { 59,  87, 175},
                        { 29,  79, 150}, { 10,  65, 118}, {  2,  47,  82},
                    },
                    {  // band 2
                        { 90, 141, 216}, { 77, 120, 210}, { 23,  95, 184},
                        { 11,  81, 151}, {  6,  75, 130}, {  2,  58, 113},
                    },
                    {  // band 3
                        {122, 167, 231}, { 66, 119, 225}, { 26,  87, 189},
                        {  7,  76, 151}, {  2,  63, 125}, {  1,  59,  77},
                    },
                    {  // band 4
                        {162, 147, 244}, {110,  97, 236}, { 32,  88, 204},
                        { 11,  89, 174}, {  5,  78, 151}, {128, 128, 128},
                    },
                    {  // band 5
                        {205,  59, 251}, {176,  68, 248}, { 90,  71, 223},
                        { 49,  72, 188}, { 17,  74, 203}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {188,  70, 207}, {140,  73, 189}, { 85,  73, 163},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 59, 144, 239}, { 79, 126, 237}, { 31, 102, 202},
                        { 10,  81, 153}, {  3,  56, 102}, {  2,  33,  59},
                    },
                    {  // band 2
                        {100, 152, 243}, { 80, 129, 236}, { 14,  94, 194},
                        {  4,  72, 150}, {  1,  50, 103}, {  1,  35,  60},
                    },
                    {  // band 3
                        {130, 183, 247}, { 70, 139, 242}, { 19, 100, 203},
                        {  4,  83, 159}, {  1,  59, 119}, {  1,  44,  72},
                    },
                    {  // band 4
                        {197, 138, 252}, {135, 107, 247}, { 31,  86, 210},
                        {  7,  74, 160}, {  1,  53, 107}, {128, 128, 128},
                    },
                    {  // band 5
                        {229,  54, 254}, {200,  51, 251}, { 83,  61, 226},
                        { 33,  55, 177}, { 12,  74, 145}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {229,  20, 235}, {183,  37, 221}, {127,  47, 198},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {188, 115, 251}, {208, 110, 250}, {101,  99, 235},
                        { 38,  81, 197}, {  9,  56, 132}, {  9,  52,  63},
                    },
                    {  // band 2
                        {189, 150, 252}, {186, 137, 251}, { 54, 107, 236},
                        { 14,  90, 195}, {  1,  89, 104}, {128, 128, 128},
                    },
                    {  // band 3
                        {209, 180, 254}, {142, 145, 253}, { 51, 130, 236},
                        {  6, 128, 214}, {  1, 128, 254}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 140, 254}, {194, 128, 254}, { 75, 119, 233},
                        {128,  23, 230}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {244,  59, 254}, {239,  81, 254}, {128,  85, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {246,  55, 247}, {197,  64, 235}, {141,  74, 218},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {178, 163, 254}, {192, 138, 252}, { 85, 103, 231},
                        { 49,  81, 179}, { 32,  54, 133}, { 12,  26,  98},
                    },
                    {  // band 2
                        {189, 173, 254}, {179, 150, 253}, { 60,  94, 237},
                        { 34,  81, 198}, { 20,  53, 187}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 191, 254}, {157, 160, 254}, { 57, 117, 240},
                        { 28, 105, 211}, {  1, 128,   1}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 146, 254}, {208, 133, 254}, { 66,  78, 233},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  49, 254}, {246,  63, 254}, { 85, 142, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 45,  28, 124}, { 23,  35, 107}, { 10,  34,  78},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 53,  99, 177}, { 82,  96, 174}, { 46,  89, 158},
                        { 21,  76, 133}, {  6,  56,  94}, {  1,  33,  54},
                    },
                    {  // band 2
                        { 68, 147, 201}, { 42, 124, 195}, { 17,  98, 166},
                        {  7,  75, 131}, {  2,  53,  93}, {  1,  33,  59},
                    },
                    {  // band 3
                        { 65, 176, 217}, { 30, 137, 206}, {  6,  97, 167},
                        {  2,  70, 128}, {  1,  47,  88}, {  1,  29,  46},
                    },
                    {  // band 4
                        { 69, 195, 232}, { 24, 146, 218}, {  4, 100, 175},
                        {  2,  72, 134}, {  1,  51,  93}, {  1,  29,  52},
                    },
                    {  // band 5
                        { 96, 212, 246}, { 39, 158, 234}, {  6, 109, 192},
                        {  2,  77, 144}, {  1,  50,  95}, {  1,  20,  45},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 71,  80, 213}, { 53,  73, 181}, { 25,  66, 141},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 168, 231}, { 91, 150, 229}, { 49, 122, 202},
                        { 22,  97, 162}, { 10,  68, 108}, {  9,  48,  57},
                    },
                    {  // band 2
                        { 56, 178, 236}, { 32, 148, 225}, {  9,  99, 176},
                        {  4,  69, 127}, {  2,  44,  78}, {  1,  25,  41},
                    },
                    {  // band 3
                        { 57, 191, 242}, { 27, 155, 230}, {  5, 102, 180},
                        {  2,  71, 133}, {  1,  44,  78}, {  1,  27,  41},
                    },
                    {  // band 4
                        { 67, 201, 247}, { 24, 162, 237}, {  3, 106, 188},
                        {  3,  74, 137}, {  1,  46,  85}, {  1,  34,  48},
                    },
                    {  // band 5
                        {111, 210, 251}, { 47, 166, 244}, {  3, 113, 199},
                        {  2,  77, 146}, {  1,  48,  93}, {  1,  38,  22},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {206,  21, 221}, {150,  36, 195}, { 94,  44, 164},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {147, 128, 239}, {194, 122, 238}, { 95, 104, 220},
                        { 39,  81, 183}, { 13,  53, 111}, {  3,  24,  49},
                    },
                    {  // band 2
                        {164, 163, 244}, {106, 142, 239}, { 50, 112, 215},
                        { 26,  90, 177}, { 12,  67, 130}, {  1,   1,  64},
                    },
                    {  // band 3
                        {155, 193, 249}, { 88, 158, 244}, { 26, 124, 220},
                        { 10,  98, 173}, {  1,  77, 126}, {128, 128, 128},
                    },
                    {  // band 4
                        {141, 205, 252}, { 64, 174, 248}, { 17, 124, 221},
                        { 12,  92, 176}, {  1,  29, 148}, {128, 128, 128},
                    },
                    {  // band 5
                        {150, 217, 254}, { 74, 191, 252}, { 30, 144, 215},
                        {  1, 106, 137}, {128,   1, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {241,  37, 242}, {175,  48, 223}, { 99,  53, 189},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {153, 183, 248}, {212, 156, 247}, {134, 124, 221},
                        { 88, 103, 184}, { 59,  86, 132}, { 29,  61,  67},
                    },
                    {  // band 2
                        {162, 199, 250}, {106, 167, 247}, { 56, 110, 207},
                        { 32,  85, 165}, { 16,  71, 130}, {  1,  93, 254},
                    },
                    {  // band 3
                        {143, 213, 252}, { 86, 187, 250}, { 23, 124, 220},
                        {  7,  95, 176}, {  1, 109, 102}, {128, 128, 128},
                    },
                    {  // band 4
                        {130, 219, 254}, { 70, 201, 253}, { 15, 128, 215},
                        {  1, 101, 201}, {  1,  64, 170}, {128, 128, 128},
                    },
                    {  // band 5
                        {155, 219, 254}, {105, 207, 254}, { 28, 155, 229},
                        {  1, 153, 191}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 18,  26, 117}, { 10,  29,  82}, {  3,  25,  52},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35,  88, 152}, { 62,  85, 150}, { 36,  77, 137},
                        { 16,  66, 116}, {  4,  47,  81}, {  1,  26,  44},
                    },
                    {  // band 2
                        { 55, 141, 182}, { 32, 119, 177}, { 12,  93, 154},
                        {  4,  71, 123}, {  1,  51,  89}, {  1,  32,  56},
                    },
                    {  // band 3
                        { 46, 171, 202}, { 21, 130, 191}, {  5,  91, 154},
                        {  1,  64, 115}, {  1,  42,  77}, {  1,  25,  41},
                    },
                    {  // band 4
                        { 43, 195, 219}, { 12, 142, 203}, {  1,  91, 156},
                        {  1,  63, 115}, {  1,  41,  77}, {  1,  22,  43},
                    },
                    {  // band 5
                        { 42, 221, 238}, {  8, 162, 219}, {  1,  98, 167},
                        {  1,  67, 123}, {  1,  43,  83}, {  1,  25,  38},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 16,  51, 216}, { 20,  48, 168}, {  9,  44, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 34, 164, 226}, {124, 148, 226}, { 72, 127, 207},
                        { 36, 107, 175}, { 15,  81, 129}, {  6,  51,  79},
                    },
                    {  // band 2
                        { 61, 182, 234}, { 35, 148, 220}, {  9, 101, 178},
                        {  4,  71, 134}, {  1,  46,  90}, {  1,  24,  51},
                    },
                    {  // band 3
                        { 54, 198, 239}, { 25, 156, 224}, {  3,  98, 173},
                        {  1,  66, 124}, {  1,  41,  78}, {  1,  15,  37},
                    },
                    {  // band 4
                        { 48, 209, 242}, { 12, 162, 226}, {  1,  96, 169},
                        {  1,  63, 119}, {  1,  40,  78}, {  1,  18,  45},
                    },
                    {  // band 5
                        { 44, 223, 247}, {  6, 173, 232}, {  1, 105, 178},
                        {  1,  71, 131}, {  1,  44,  84}, {  1,  13,  46},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {188,  26, 214}, {121,  42, 181}, { 66,  49, 149},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {136, 128, 233}, {172, 124, 230}, { 80, 106, 211},
                        { 27,  81, 174}, {  6,  49,  98}, {  8,  28,  49},
                    },
                    {  // band 2
                        {145, 166, 239}, { 92, 141, 229}, { 28, 108, 196},
                        {  8,  87, 154}, {  1,  58, 105}, {  1,  27,  59},
                    },
                    {  // band 3
                        {131, 193, 242}, { 66, 151, 231}, { 13, 112, 192},
                        {  2,  81, 152}, {  1,  66, 121}, {  1,  23,  64},
                    },
                    {  // band 4
                        {112, 211, 246}, { 41, 164, 235}, {  5, 117, 202},
                        {  1,  83, 162}, {  1,  64, 111}, {128, 128, 128},
                    },
                    {  // band 5
                        { 96, 230, 250}, { 28, 185, 243}, {  2, 132, 204},
                        {  1,  91, 166}, {  1,  85,  46}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {238,  23, 242}, {157,  29, 215}, { 73,  27, 162},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {165, 173, 250}, {222, 151, 247}, {152, 134, 235},
                        {114, 120, 210}, { 86, 109, 176}, { 53,  88, 145},
                    },
                    {  // band 2
                        {164, 194, 249}, {100, 158, 241}, { 35, 111, 212},
                        { 17,  85, 167}, {  1,  52, 112}, {  1,  73,   1},
                    },
                    {  // band 3
                        {151, 215, 252}, { 83, 172, 245}, { 16, 122, 208},
                        {  6, 101, 165}, {  1,  74, 113}, {  1,   1,   1},
                    },
                    {  // band 4
                        {138, 230, 253}, { 65, 184, 248}, {  8, 128, 212},
                        {  1, 111, 182}, {128,   1,   1}, {128, 128, 128},
                    },
                    {  // band 5
                        {123, 240, 253}, { 36, 201, 250}, {  3, 127, 211},
                        {  1,  68, 204}, {128,   1,   1}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 51,  21, 156}, { 30,  23,  86}, {  4,  18,  37},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 38,  77, 129}, { 79,  76, 129}, { 40,  66, 117},
                        { 12,  54,  95}, {  1,  36,  60}, {  1,  17,  29},
                    },
                    {  // band 2
                        { 44, 133, 149}, { 24, 107, 143}, {  8,  78, 121},
                        {  3,  59,  97}, {  1,  42,  71}, {  1,  22,  37},
                    },
                    {  // band 3
                        { 29, 160, 171}, {  9, 114, 158}, {  1,  76, 125},
                        {  1,  54,  93}, {  1,  36,  63}, {  1,  20,  35},
                    },
                    {  // band 4
                        { 22, 188, 205}, {  6, 132, 186}, {  1,  87, 144},
                        {  1,  62, 107}, {  1,  41,  72}, {  1,  23,  41},
                    },
                    {  // band 5
                        { 25, 233, 236}, {  5, 165, 214}, {  1,  96, 158},
                        {  1,  63, 112}, {  1,  40,  73}, {  1,  23,  40},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 48,  20, 231}, { 37,  21, 179}, { 15,  18, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41, 154, 216}, {196, 142, 221}, {131, 125, 207},
                        { 84, 111, 181}, { 45,  91, 142}, { 27,  62,  89},
                    },
                    {  // band 2
                        { 72, 181, 230}, { 41, 147, 215}, { 10, 102, 173},
                        {  3,  73, 132}, {  1,  47,  89}, {  1,  23,  50},
                    },
                    {  // band 3
                        { 60, 201, 236}, { 23, 157, 219}, {  2,  99, 167},
                        {  1,  69, 124}, {  1,  43,  80}, {  1,  22,  39},
                    },
                    {  // band 4
                        { 53, 214, 242}, { 15, 165, 224}, {  1, 101, 173},
                        {  1,  70, 131}, {  1,  44,  83}, {  1,  23,  49},
                    },
                    {  // band 5
                        { 39, 239, 248}, {  7, 186, 233}, {  1, 108, 174},
                        {  1,  70, 123}, {  1,  43,  77}, {  1,  16,  42},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {161,  26, 204}, { 77,  40, 160}, { 26,  50, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80, 140, 218}, {136, 133, 215}, { 63, 117, 197},
                        { 20,  93, 170}, {  7,  55, 102}, { 13,  32,  52},
                    },
                    {  // band 2
                        { 86, 173, 231}, { 46, 150, 220}, { 18, 118, 190},
                        {  8,  90, 150}, {  2,  60,  95}, {  1,  39,  41},
                    },
                    {  // band 3
                        { 80, 183, 242}, { 37, 160, 231}, {  6, 120, 182},
                        {  1,  86, 137}, {  1,  46,  78}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 88, 215, 247}, { 42, 179, 235}, {  4, 116, 182},
                        {  2,  80, 133}, {  1,  46,  85}, {  1,  64,  43},
                    },
                    {  // band 5
                        {100, 236, 250}, { 31, 186, 234}, {  1, 114, 181},
                        {  1,  85, 135}, {  1,  78,  64}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  13, 245}, {106,  16, 211}, { 32,  11, 156},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {140, 214, 247}, {241, 186, 243}, {177, 172, 235},
                        {128, 156, 219}, {106, 130, 191}, { 99, 105, 152},
                    },
                    {  // band 2
                        {125, 218, 248}, { 75, 167, 239}, { 29, 111, 212},
                        {  6,  66, 152}, {  1,  42,  96}, {  1,  85, 128},
                    },
                    {  // band 3
                        {120, 232, 252}, { 60, 189, 247}, {  8, 141, 200},
                        {  1,  89, 134}, {  1,  32, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {111, 238, 253}, { 56, 198, 245}, {  1, 123, 208},
                        {  1,  93, 176}, {  1,   1,  73}, {128, 128, 128},
                    },
                    {  // band 5
                        { 98, 251, 249}, { 56, 189, 244}, { 17, 113, 220},
                        {  1, 109, 179}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 51,  21, 156}, { 30,  23,  86}, {  4,  18,  37},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 38,  77, 129}, { 79,  76, 129}, { 40,  66, 117},
                        { 12,  54,  95}, {  1,  36,  60}, {  1,  17,  29},
                    },
                    {  // band 2
                        { 44, 133, 149}, { 24, 107, 143}, {  8,  78, 121},
                        {  3,  59,  97}, {  1,  42,  71}, {  1,  22,  37},
                    },
                    {  // band 3
                        { 29, 160, 171}, {  9, 114, 158}, {  1,  76, 125},
                        {  1,  54,  93}, {  1,  36,  63}, {  1,  20,  35},
                    },
                    {  // band 4
                        { 22, 188, 205}, {  6, 132, 186}, {  1,  87, 144},
                        {  1,  62, 107}, {  1,  41,  72}, {  1,  23,  41},
                    },
                    {  // band 5
                        { 25, 233, 236}, {  5, 165, 214}, {  1,  96, 158},
                        {  1,  63, 112}, {  1,  40,  73}, {  1,  23,  40},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 48,  20, 231}, { 37,  21, 179}, { 15,  18, 109},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 41, 154, 216}, {196, 142, 221}, {131, 125, 207},
                        { 84, 111, 181}, { 45,  91, 142}, { 27,  62,  89},
                    },
                    {  // band 2
                        { 72, 181, 230}, { 41, 147, 215}, { 10, 102, 173},
                        {  3,  73, 132}, {  1,  47,  89}, {  1,  23,  50},
                    },
                    {  // band 3
                        { 60, 201, 236}, { 23, 157, 219}, {  2,  99, 167},
                        {  1,  69, 124}, {  1,  43,  80}, {  1,  22,  39},
                    },
                    {  // band 4
                        { 53, 214, 242}, { 15, 165, 224}, {  1, 101, 173},
                        {  1,  70, 131}, {  1,  44,  83}, {  1,  23,  49},
                    },
                    {  // band 5
                        { 39, 239, 248}, {  7, 186, 233}, {  1, 108, 174},
                        {  1,  70, 123}, {  1,  43,  77}, {  1,  16,  42},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {161,  26, 204}, { 77,  40, 160}, { 26,  50, 117},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 80, 140, 218}, {136, 133, 215}, { 63, 117, 197},
                        { 20,  93, 170}, {  7,  55, 102}, { 13,  32,  52},
                    },
                    {  // band 2
                        { 86, 173, 231}, { 46, 150, 220}, { 18, 118, 190},
                        {  8,  90, 150}, {  2,  60,  95}, {  1,  39,  41},
                    },
                    {  // band 3
                        { 80, 183, 242}, { 37, 160, 231}, {  6, 120, 182},
                        {  1,  86, 137}, {  1,  46,  78}, {  1,  15,  24},
                    },
                    {  // band 4
                        { 88, 215, 247}, { 42, 179, 235}, {  4, 116, 182},
                        {  2,  80, 133}, {  1,  46,  85}, {  1,  64,  43},
                    },
                    {  // band 5
                        {100, 236, 250}, { 31, 186, 234}, {  1, 114, 181},
                        {  1,  85, 135}, {  1,  78,  64}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {213,  13, 245}, {106,  16, 211}, { 32,  11, 156},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {140, 214, 247}, {241, 186, 243}, {177, 172, 235},
                        {128, 156, 219}, {106, 130, 191}, { 99, 105, 152},
                    },
                    {  // band 2
                        {125, 218, 248}, { 75, 167, 239}, { 29, 111, 212},
                        {  6,  66, 152}, {  1,  42,  96}, {  1,  85, 128},
                    },
                    {  // band 3
                        {120, 232, 252}, { 60, 189, 247}, {  8, 141, 200},
                        {  1,  89, 134}, {  1,  32, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {111, 238, 253}, { 56, 198, 245}, {  1, 123, 208},
                        {  1,  93, 176}, {  1,   1,  73}, {128, 128, 128},
                    },
                    {  // band 5
                        { 98, 251, 249}, { 56, 189, 244}, { 17, 113, 220},
                        {  1, 109, 179}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
    {  // Q_Index 3
        {  // TX_SIZE 0
            {  // Y plane
                {  // Intra
                    {  // band 0
                        {186,  16, 200}, {122,  31, 187}, { 78,  40, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {146, 119, 245}, {182, 115, 244}, {130, 113, 238},
                        { 88, 110, 225}, { 47, 103, 208}, {  5, 102, 188},
                    },
                    {  // band 2
                        {164, 157, 248}, {155, 141, 250}, { 71, 116, 243},
                        { 88, 129, 233}, { 50,  99, 228}, { 26, 148, 191},
                    },
                    {  // band 3
                        {200, 158, 253}, {177, 118, 252}, { 99, 113, 245},
                        { 77, 120, 210}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {231, 104, 254}, {209,  82, 254}, {143, 112, 252},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {250,  36, 254}, {243,  55, 254}, {223, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        {207,  37, 226}, {164,  46, 218}, {122,  58, 201},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {149, 154, 253}, {170, 137, 253}, { 94, 123, 247},
                        { 42, 113, 222}, { 16,  97, 174}, { 49,  98, 159},
                    },
                    {  // band 2
                        {177, 162, 253}, {165, 142, 252}, { 51, 108, 243},
                        { 18, 108, 213}, {  1,  98, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {211, 152, 254}, {184, 116, 254}, { 70, 110, 244},
                        {  8, 108, 237}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {236,  89, 254}, {210,  67, 254}, {112, 111, 248},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {246,  26, 254}, {233,  35, 254}, {128,   1, 254},
                        {254, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {247,   2, 247}, {226,   8, 242}, {191,  14, 235},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {231,  94, 254}, {248,  91, 254}, {186,  89, 252},
                        {128,  92, 244}, { 79, 112, 254}, {128, 128, 128},
                    },
                    {  // band 2
                        {228, 145, 253}, {240, 130, 254}, {223, 105, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {245, 153, 253}, {240, 120, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 128, 254}, {204, 128, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {253,   7, 249}, {224,   9, 244}, {182,  13, 231},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {234, 109, 254}, {242, 104, 254}, {160,  98, 254},
                        {123,  85, 243}, { 82,  43, 217}, {128, 128, 128},
                    },
                    {  // band 2
                        {243, 137, 254}, {240, 118, 254}, {136,  53, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {251, 173, 254}, {229, 129, 250}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {254, 119, 254}, {254, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 1
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 49,  26, 159}, { 36,  34, 150}, { 26,  38, 124},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 99, 122, 226}, {143, 119, 225}, { 90, 113, 213},
                        { 46, 102, 193}, { 14,  84, 157}, {  3,  59, 107},
                    },
                    {  // band 2
                        {109, 164, 237}, { 74, 142, 233}, { 29, 112, 216},
                        { 14,  92, 184}, { 10,  80, 156}, {  1,  52, 137},
                    },
                    {  // band 3
                        {110, 191, 245}, { 59, 156, 240}, { 18, 121, 220},
                        {  8,  97, 184}, {  3,  84, 150}, {128, 128, 128},
                    },
                    {  // band 4
                        {115, 203, 250}, { 59, 167, 246}, { 16, 130, 226},
                        {  7,  97, 192}, {  1,  71,  99}, {128, 128, 128},
                    },
                    {  // band 5
                        {149, 218, 253}, { 93, 171, 251}, { 28, 125, 233},
                        { 28,  99, 192}, {128,  85,  85}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 97,  45, 229}, { 79,  52, 205}, { 46,  58, 171},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 99, 180, 249}, {156, 165, 249}, { 73, 141, 237},
                        { 31, 116, 208}, { 13,  81, 153}, {  5,  42,  86},
                    },
                    {  // band 2
                        {113, 188, 251}, { 68, 161, 244}, { 16, 108, 216},
                        {  6,  81, 168}, {  2,  65, 118}, {128,   1,   1},
                    },
                    {  // band 3
                        {117, 201, 252}, { 62, 171, 248}, { 12, 119, 221},
                        {  5,  90, 182}, {  4,  66, 116}, {128, 128, 128},
                    },
                    {  // band 4
                        {128, 207, 253}, { 70, 176, 251}, { 11, 126, 228},
                        {  6,  89, 189}, {  1,  44, 148}, {128, 128, 128},
                    },
                    {  // band 5
                        {162, 218, 254}, {107, 170, 253}, { 22, 131, 238},
                        {  1,  77, 182}, {  1, 254, 128}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {235,   5, 238}, {194,  14, 223}, {152,  22, 205},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {200, 121, 251}, {241, 115, 252}, {167, 108, 248},
                        { 93,  93, 233}, { 36,  66, 189}, {128, 128, 128},
                    },
                    {  // band 2
                        {220, 151, 253}, {176, 135, 252}, { 95, 124, 254},
                        { 64, 105, 217}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {225, 189, 254}, {175, 155, 254}, {102, 119, 254},
                        {  1,   1,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {218, 195, 254}, {125, 157, 253}, {128, 128, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {221, 197, 254}, { 85, 210, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {250,   9, 246}, {204,  13, 234}, {144,  18, 211},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {213, 157, 253}, {243, 138, 253}, {170, 117, 250},
                        {109,  91, 233}, { 66,  77, 163}, { 64,  85, 254},
                    },
                    {  // band 2
                        {221, 169, 254}, {182, 141, 253}, {112, 120, 239},
                        { 85, 165, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {226, 192, 254}, {189, 174, 251}, {153, 128, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {232, 192, 254}, {195, 187, 247}, {  1, 191, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {247, 185, 254}, {254,  93, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 2
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 14,  30, 136}, { 15,  33, 120}, { 10,  33,  90},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 92, 109, 209}, {113, 108, 207}, { 77, 102, 193},
                        { 39,  91, 171}, { 11,  70, 129}, {  2,  44,  77},
                    },
                    {  // band 2
                        { 99, 158, 223}, { 66, 135, 217}, { 23, 109, 194},
                        {  9,  85, 160}, {  3,  66, 124}, {  1,  51, 100},
                    },
                    {  // band 3
                        { 89, 189, 234}, { 46, 149, 225}, { 10, 110, 194},
                        {  2,  83, 156}, {  1,  57, 113}, {  1,  47,  73},
                    },
                    {  // band 4
                        { 78, 206, 242}, { 28, 161, 232}, {  3, 114, 200},
                        {  1,  86, 161}, {  1,  62, 118}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 72, 227, 250}, { 20, 182, 242}, {  3, 126, 210},
                        {  2,  91, 166}, {  1,  64, 126}, {128, 128, 128},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 23,  42, 227}, { 41,  43, 195}, { 25,  45, 146},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {100, 172, 245}, {165, 158, 246}, { 88, 137, 234},
                        { 44, 116, 203}, { 18,  85, 149}, {  7,  56,  92},
                    },
                    {  // band 2
                        {117, 188, 247}, { 70, 155, 239}, { 18, 105, 204},
                        {  7,  78, 158}, {  2,  50, 111}, {  1,  38,  77},
                    },
                    {  // band 3
                        {104, 207, 250}, { 54, 166, 241}, {  6, 110, 199},
                        {  1,  78, 155}, {  1,  45, 100}, {  1,   1,   1},
                    },
                    {  // band 4
                        { 87, 216, 251}, { 30, 177, 243}, {  1, 114, 203},
                        {  1,  85, 157}, {  1,  53, 108}, {128, 128, 128},
                    },
                    {  // band 5
                        { 80, 230, 253}, { 23, 193, 248}, {  1, 127, 215},
                        {  1,  94, 170}, {  1,  71,  59}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {222,   9, 234}, {161,  20, 210}, {113,  30, 185},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {195, 120, 248}, {231, 124, 247}, {148, 116, 238},
                        { 64,  98, 207}, { 20,  70, 147}, { 87,  68, 100},
                    },
                    {  // band 2
                        {186, 161, 250}, {124, 148, 245}, { 44, 123, 230},
                        { 23, 107, 205}, {  1,  80, 131}, {128, 128, 128},
                    },
                    {  // band 3
                        {172, 196, 252}, {110, 160, 248}, { 37, 134, 235},
                        { 23, 125, 200}, {128, 254, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {173, 209, 253}, {103, 175, 250}, {  1, 120, 240},
                        {  1, 146, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {184, 235, 254}, { 81, 186, 251}, {128, 109, 254},
                        {128, 254, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {248,   8, 243}, {185,  11, 225}, {108,  11, 189},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {208, 158, 254}, {244, 147, 252}, {195, 132, 248},
                        {161, 122, 224}, {129, 114, 188}, { 59, 119, 159},
                    },
                    {  // band 2
                        {202, 182, 253}, {143, 161, 251}, { 73, 115, 247},
                        {146, 175, 204}, {128,   1, 254}, {128, 128, 128},
                    },
                    {  // band 3
                        {202, 204, 254}, {131, 174, 251}, { 18, 153, 207},
                        {128, 254, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {192, 221, 254}, {114, 190, 254}, {128, 170, 254},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {166, 236, 254}, {119, 200, 254}, {128, 128, 128},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
        {  // TX_SIZE 3
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 30,  32, 144}, { 21,  35,  96}, {  4,  27,  55},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 107, 172}, { 61, 104, 170}, { 33,  94, 160},
                        { 13,  80, 139}, {  2,  55,  97}, {  1,  28,  49},
                    },
                    {  // band 2
                        { 51, 153, 195}, { 29, 129, 189}, {  9,  99, 163},
                        {  3,  75, 129}, {  1,  49,  88}, {  1,  29,  50},
                    },
                    {  // band 3
                        { 53, 164, 210}, { 21, 134, 201}, {  3,  97, 164},
                        {  1,  69, 124}, {  1,  45,  82}, {  1,  31,  58},
                    },
                    {  // band 4
                        { 47, 205, 234}, { 18, 158, 220}, {  2, 109, 177},
                        {  1,  78, 137}, {  1,  53, 101}, {  1,  34,  70},
                    },
                    {  // band 5
                        { 55, 233, 245}, { 16, 179, 233}, {  1, 116, 191},
                        {  1,  79, 145}, {  1,  53, 101}, {  1,  37,  58},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 36,  33, 227}, { 39,  28, 190}, { 18,  27, 134},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 156, 235}, {184, 147, 235}, {114, 130, 220},
                        { 72, 112, 191}, { 42,  87, 144}, { 21,  65,  93},
                    },
                    {  // band 2
                        { 96, 179, 240}, { 51, 149, 228}, { 12, 105, 191},
                        {  6,  74, 148}, {  1,  47, 100}, {  1,  29,  53},
                    },
                    {  // band 3
                        { 88, 191, 242}, { 35, 154, 231}, {  3, 106, 187},
                        {  1,  74, 140}, {  1,  41,  84}, {  1,  25,  38},
                    },
                    {  // band 4
                        { 77, 212, 249}, { 28, 171, 239}, {  2, 117, 199},
                        {  1,  79, 151}, {  1,  45,  99}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 77, 236, 252}, { 27, 190, 246}, {  2, 120, 203},
                        {  1,  78, 147}, {  1,  42,  72}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {185,  11, 227}, {113,  30, 182}, { 57,  44, 144},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {151, 139, 244}, {212, 139, 241}, {124, 126, 231},
                        { 59, 104, 213}, { 26,  73, 158}, { 20,  45,  95},
                    },
                    {  // band 2
                        {155, 163, 247}, {108, 152, 239}, { 39, 124, 214},
                        {  7, 109, 162}, { 29,  57, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {158, 176, 250}, { 89, 164, 243}, { 11, 114, 196},
                        {  1,  96, 141}, {  1,  81, 118}, {128,   1,   1},
                    },
                    {  // band 4
                        {148, 212, 251}, { 59, 174, 240}, {  2, 130, 203},
                        {  1,  70, 168}, {  1,  51, 106}, {128, 128, 128},
                    },
                    {  // band 5
                        {104, 237, 252}, { 39, 190, 246}, {  1, 154, 220},
                        {128, 102,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {236,   6, 242}, {111,   6, 206}, { 36,   5, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {193, 193, 252}, {248, 182, 251}, {218, 150, 246},
                        {182, 134, 244}, {151, 137, 227}, { 45, 102, 195},
                    },
                    {  // band 2
                        {188, 202, 251}, {125, 165, 249}, { 64,  75, 218},
                        {  1, 128, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {178, 225, 254}, {107, 188, 231}, { 21, 135, 233},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {164, 227, 253}, { 55, 193, 251}, {  1, 111, 225},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {151, 243, 254}, { 50, 203, 254}, {128, 179, 254},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#if CONFIG_TX64X64
        {  // TX_SIZE 4
            {  // Y plane
                {  // Intra
                    {  // band 0
                        { 30,  32, 144}, { 21,  35,  96}, {  4,  27,  55},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 35, 107, 172}, { 61, 104, 170}, { 33,  94, 160},
                        { 13,  80, 139}, {  2,  55,  97}, {  1,  28,  49},
                    },
                    {  // band 2
                        { 51, 153, 195}, { 29, 129, 189}, {  9,  99, 163},
                        {  3,  75, 129}, {  1,  49,  88}, {  1,  29,  50},
                    },
                    {  // band 3
                        { 53, 164, 210}, { 21, 134, 201}, {  3,  97, 164},
                        {  1,  69, 124}, {  1,  45,  82}, {  1,  31,  58},
                    },
                    {  // band 4
                        { 47, 205, 234}, { 18, 158, 220}, {  2, 109, 177},
                        {  1,  78, 137}, {  1,  53, 101}, {  1,  34,  70},
                    },
                    {  // band 5
                        { 55, 233, 245}, { 16, 179, 233}, {  1, 116, 191},
                        {  1,  79, 145}, {  1,  53, 101}, {  1,  37,  58},
                    },
                },
                {  // Intra
                    {  // band 0
                        { 36,  33, 227}, { 39,  28, 190}, { 18,  27, 134},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        { 76, 156, 235}, {184, 147, 235}, {114, 130, 220},
                        { 72, 112, 191}, { 42,  87, 144}, { 21,  65,  93},
                    },
                    {  // band 2
                        { 96, 179, 240}, { 51, 149, 228}, { 12, 105, 191},
                        {  6,  74, 148}, {  1,  47, 100}, {  1,  29,  53},
                    },
                    {  // band 3
                        { 88, 191, 242}, { 35, 154, 231}, {  3, 106, 187},
                        {  1,  74, 140}, {  1,  41,  84}, {  1,  25,  38},
                    },
                    {  // band 4
                        { 77, 212, 249}, { 28, 171, 239}, {  2, 117, 199},
                        {  1,  79, 151}, {  1,  45,  99}, {  1,   1,   1},
                    },
                    {  // band 5
                        { 77, 236, 252}, { 27, 190, 246}, {  2, 120, 203},
                        {  1,  78, 147}, {  1,  42,  72}, {128, 128, 128},
                    },
                },
            },
            {  // UV plane
                {  // Inter
                    {  // band 0
                        {185,  11, 227}, {113,  30, 182}, { 57,  44, 144},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {151, 139, 244}, {212, 139, 241}, {124, 126, 231},
                        { 59, 104, 213}, { 26,  73, 158}, { 20,  45,  95},
                    },
                    {  // band 2
                        {155, 163, 247}, {108, 152, 239}, { 39, 124, 214},
                        {  7, 109, 162}, { 29,  57, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {158, 176, 250}, { 89, 164, 243}, { 11, 114, 196},
                        {  1,  96, 141}, {  1,  81, 118}, {128,   1,   1},
                    },
                    {  // band 4
                        {148, 212, 251}, { 59, 174, 240}, {  2, 130, 203},
                        {  1,  70, 168}, {  1,  51, 106}, {128, 128, 128},
                    },
                    {  // band 5
                        {104, 237, 252}, { 39, 190, 246}, {  1, 154, 220},
                        {128, 102,   1}, {128, 128, 128}, {128, 128, 128},
                    },
                },
                {  // Inter
                    {  // band 0
                        {236,   6, 242}, {111,   6, 206}, { 36,   5, 161},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 1
                        {193, 193, 252}, {248, 182, 251}, {218, 150, 246},
                        {182, 134, 244}, {151, 137, 227}, { 45, 102, 195},
                    },
                    {  // band 2
                        {188, 202, 251}, {125, 165, 249}, { 64,  75, 218},
                        {  1, 128, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 3
                        {178, 225, 254}, {107, 188, 231}, { 21, 135, 233},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 4
                        {164, 227, 253}, { 55, 193, 251}, {  1, 111, 225},
                        {128, 128, 128}, {128, 128, 128}, {128, 128, 128},
                    },
                    {  // band 5
                        {151, 243, 254}, { 50, 203, 254}, {128, 179, 254},
                        {128,   1, 254}, {128, 128, 128}, {128, 128, 128},
                    },
                },
            },
        },
#endif  // CONFIG_TX64X64
    },
};
#else
static const av1_coeff_probs_model default_coef_probs_4x4[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        { 195,  29, 183 }, {  84,  49, 136 }, {   8,  42,  71 }
      }, {  // Band 1
        {  31, 107, 169 }, {  35,  99, 159 }, {  17,  82, 140 },
        {   8,  66, 114 }, {   2,  44,  76 }, {   1,  19,  32 }
      }, {  // Band 2
        {  40, 132, 201 }, {  29, 114, 187 }, {  13,  91, 157 },
        {   7,  75, 127 }, {   3,  58,  95 }, {   1,  28,  47 }
      }, {  // Band 3
        {  69, 142, 221 }, {  42, 122, 201 }, {  15,  91, 159 },
        {   6,  67, 121 }, {   1,  42,  77 }, {   1,  17,  31 }
      }, {  // Band 4
        { 102, 148, 228 }, {  67, 117, 204 }, {  17,  82, 154 },
        {   6,  59, 114 }, {   2,  39,  75 }, {   1,  15,  29 }
      }, {  // Band 5
        { 156,  57, 233 }, { 119,  57, 212 }, {  58,  48, 163 },
        {  29,  40, 124 }, {  12,  30,  81 }, {   3,  12,  31 }
      }
    }, {  // Inter
      {  // Band 0
        { 191, 107, 226 }, { 124, 117, 204 }, {  25,  99, 155 }
      }, {  // Band 1
        {  29, 148, 210 }, {  37, 126, 194 }, {   8,  93, 157 },
        {   2,  68, 118 }, {   1,  39,  69 }, {   1,  17,  33 }
      }, {  // Band 2
        {  41, 151, 213 }, {  27, 123, 193 }, {   3,  82, 144 },
        {   1,  58, 105 }, {   1,  32,  60 }, {   1,  13,  26 }
      }, {  // Band 3
        {  59, 159, 220 }, {  23, 126, 198 }, {   4,  88, 151 },
        {   1,  66, 114 }, {   1,  38,  71 }, {   1,  18,  34 }
      }, {  // Band 4
        { 114, 136, 232 }, {  51, 114, 207 }, {  11,  83, 155 },
        {   3,  56, 105 }, {   1,  33,  65 }, {   1,  17,  34 }
      }, {  // Band 5
        { 149,  65, 234 }, { 121,  57, 215 }, {  61,  49, 166 },
        {  28,  36, 114 }, {  12,  25,  76 }, {   3,  16,  42 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 214,  49, 220 }, { 132,  63, 188 }, {  42,  65, 137 }
      }, {  // Band 1
        {  85, 137, 221 }, { 104, 131, 216 }, {  49, 111, 192 },
        {  21,  87, 155 }, {   2,  49,  87 }, {   1,  16,  28 }
      }, {  // Band 2
        {  89, 163, 230 }, {  90, 137, 220 }, {  29, 100, 183 },
        {  10,  70, 135 }, {   2,  42,  81 }, {   1,  17,  33 }
      }, {  // Band 3
        { 108, 167, 237 }, {  55, 133, 222 }, {  15,  97, 179 },
        {   4,  72, 135 }, {   1,  45,  85 }, {   1,  19,  38 }
      }, {  // Band 4
        { 124, 146, 240 }, {  66, 124, 224 }, {  17,  88, 175 },
        {   4,  58, 122 }, {   1,  36,  75 }, {   1,  18,  37 }
      }, {  //  Band 5
        { 141,  79, 241 }, { 126,  70, 227 }, {  66,  58, 182 },
        {  30,  44, 136 }, {  12,  34,  96 }, {   2,  20,  47 }
      }
    }, {  // Inter
      {  // Band 0
        { 229,  99, 249 }, { 143, 111, 235 }, {  46, 109, 192 }
      }, {  // Band 1
        {  82, 158, 236 }, {  94, 146, 224 }, {  25, 117, 191 },
        {   9,  87, 149 }, {   3,  56,  99 }, {   1,  33,  57 }
      }, {  // Band 2
        {  83, 167, 237 }, {  68, 145, 222 }, {  10, 103, 177 },
        {   2,  72, 131 }, {   1,  41,  79 }, {   1,  20,  39 }
      }, {  // Band 3
        {  99, 167, 239 }, {  47, 141, 224 }, {  10, 104, 178 },
        {   2,  73, 133 }, {   1,  44,  85 }, {   1,  22,  47 }
      }, {  // Band 4
        { 127, 145, 243 }, {  71, 129, 228 }, {  17,  93, 177 },
        {   3,  61, 124 }, {   1,  41,  84 }, {   1,  21,  52 }
      }, {  // Band 5
        { 157,  78, 244 }, { 140,  72, 231 }, {  69,  58, 184 },
        {  31,  44, 137 }, {  14,  38, 105 }, {   8,  23,  61 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_8x8[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        { 125,  34, 187 }, {  52,  41, 133 }, {   6,  31,  56 }
      }, {  // Band 1
        {  37, 109, 153 }, {  51, 102, 147 }, {  23,  87, 128 },
        {   8,  67, 101 }, {   1,  41,  63 }, {   1,  19,  29 }
      }, {  // Band 2
        {  31, 154, 185 }, {  17, 127, 175 }, {   6,  96, 145 },
        {   2,  73, 114 }, {   1,  51,  82 }, {   1,  28,  45 }
      }, {  // Band 3
        {  23, 163, 200 }, {  10, 131, 185 }, {   2,  93, 148 },
        {   1,  67, 111 }, {   1,  41,  69 }, {   1,  14,  24 }
      }, {  // Band 4
        {  29, 176, 217 }, {  12, 145, 201 }, {   3, 101, 156 },
        {   1,  69, 111 }, {   1,  39,  63 }, {   1,  14,  23 }
      }, {  // Band 5
        {  57, 192, 233 }, {  25, 154, 215 }, {   6, 109, 167 },
        {   3,  78, 118 }, {   1,  48,  69 }, {   1,  21,  29 }
      }
    }, {  // Inter
      {  // Band 0
        { 202, 105, 245 }, { 108, 106, 216 }, {  18,  90, 144 }
      }, {  // Band 1
        {  33, 172, 219 }, {  64, 149, 206 }, {  14, 117, 177 },
        {   5,  90, 141 }, {   2,  61,  95 }, {   1,  37,  57 }
      }, {  // Band 2
        {  33, 179, 220 }, {  11, 140, 198 }, {   1,  89, 148 },
        {   1,  60, 104 }, {   1,  33,  57 }, {   1,  12,  21 }
      }, {  // Band 3
        {  30, 181, 221 }, {   8, 141, 198 }, {   1,  87, 145 },
        {   1,  58, 100 }, {   1,  31,  55 }, {   1,  12,  20 }
      }, {  // Band 4
        {  32, 186, 224 }, {   7, 142, 198 }, {   1,  86, 143 },
        {   1,  58, 100 }, {   1,  31,  55 }, {   1,  12,  22 }
      }, {  // Band 5
        {  57, 192, 227 }, {  20, 143, 204 }, {   3,  96, 154 },
        {   1,  68, 112 }, {   1,  42,  69 }, {   1,  19,  32 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 212,  35, 215 }, { 113,  47, 169 }, {  29,  48, 105 }
      }, {  // Band 1
        {  74, 129, 203 }, { 106, 120, 203 }, {  49, 107, 178 },
        {  19,  84, 144 }, {   4,  50,  84 }, {   1,  15,  25 }
      }, {  // Band 2
        {  71, 172, 217 }, {  44, 141, 209 }, {  15, 102, 173 },
        {   6,  76, 133 }, {   2,  51,  89 }, {   1,  24,  42 }
      }, {  // Band 3
        {  64, 185, 231 }, {  31, 148, 216 }, {   8, 103, 175 },
        {   3,  74, 131 }, {   1,  46,  81 }, {   1,  18,  30 }
      }, {  // Band 4
        {  65, 196, 235 }, {  25, 157, 221 }, {   5, 105, 174 },
        {   1,  67, 120 }, {   1,  38,  69 }, {   1,  15,  30 }
      }, {  // Band 5
        {  65, 204, 238 }, {  30, 156, 224 }, {   7, 107, 177 },
        {   2,  70, 124 }, {   1,  42,  73 }, {   1,  18,  34 }
      }
    }, {  // Inter
      {  // Band 0
        { 225,  86, 251 }, { 144, 104, 235 }, {  42,  99, 181 }
      }, {  // Band 1
        {  85, 175, 239 }, { 112, 165, 229 }, {  29, 136, 200 },
        {  12, 103, 162 }, {   6,  77, 123 }, {   2,  53,  84 }
      }, {  // Band 2
        {  75, 183, 239 }, {  30, 155, 221 }, {   3, 106, 171 },
        {   1,  74, 128 }, {   1,  44,  76 }, {   1,  17,  28 }
      }, {  // Band 3
        {  73, 185, 240 }, {  27, 159, 222 }, {   2, 107, 172 },
        {   1,  75, 127 }, {   1,  42,  73 }, {   1,  17,  29 }
      }, {  // Band 4
        {  62, 190, 238 }, {  21, 159, 222 }, {   2, 107, 172 },
        {   1,  72, 122 }, {   1,  40,  71 }, {   1,  18,  32 }
      }, {  // Band 5
        {  61, 199, 240 }, {  27, 161, 226 }, {   4, 113, 180 },
        {   1,  76, 129 }, {   1,  46,  80 }, {   1,  23,  41 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_16x16[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {   7,  27, 153 }, {   5,  30,  95 }, {   1,  16,  30 }
      }, {  // Band 1
        {  50,  75, 127 }, {  57,  75, 124 }, {  27,  67, 108 },
        {  10,  54,  86 }, {   1,  33,  52 }, {   1,  12,  18 }
      }, {  // Band 2
        {  43, 125, 151 }, {  26, 108, 148 }, {   7,  83, 122 },
        {   2,  59,  89 }, {   1,  38,  60 }, {   1,  17,  27 }
      }, {  // Band 3
        {  23, 144, 163 }, {  13, 112, 154 }, {   2,  75, 117 },
        {   1,  50,  81 }, {   1,  31,  51 }, {   1,  14,  23 }
      }, {  // Band 4
        {  18, 162, 185 }, {   6, 123, 171 }, {   1,  78, 125 },
        {   1,  51,  86 }, {   1,  31,  54 }, {   1,  14,  23 }
      }, {  // Band 5
        {  15, 199, 227 }, {   3, 150, 204 }, {   1,  91, 146 },
        {   1,  55,  95 }, {   1,  30,  53 }, {   1,  11,  20 }
      }
    }, {  // Inter
      {  // Band 0
        {  19,  55, 240 }, {  19,  59, 196 }, {   3,  52, 105 }
      }, {  // Band 1
        {  41, 166, 207 }, { 104, 153, 199 }, {  31, 123, 181 },
        {  14, 101, 152 }, {   5,  72, 106 }, {   1,  36,  52 }
      }, {  // Band 2
        {  35, 176, 211 }, {  12, 131, 190 }, {   2,  88, 144 },
        {   1,  60, 101 }, {   1,  36,  60 }, {   1,  16,  28 }
      }, {  // Band 3
        {  28, 183, 213 }, {   8, 134, 191 }, {   1,  86, 142 },
        {   1,  56,  96 }, {   1,  30,  53 }, {   1,  12,  20 }
      }, {  // Band 4
        {  20, 190, 215 }, {   4, 135, 192 }, {   1,  84, 139 },
        {   1,  53,  91 }, {   1,  28,  49 }, {   1,  11,  20 }
      }, {  // Band 5
        {  13, 196, 216 }, {   2, 137, 192 }, {   1,  86, 143 },
        {   1,  57,  99 }, {   1,  32,  56 }, {   1,  13,  24 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 211,  29, 217 }, {  96,  47, 156 }, {  22,  43,  87 }
      }, {  // Band 1
        {  78, 120, 193 }, { 111, 116, 186 }, {  46, 102, 164 },
        {  15,  80, 128 }, {   2,  49,  76 }, {   1,  18,  28 }
      }, {  // Band 2
        {  71, 161, 203 }, {  42, 132, 192 }, {  10,  98, 150 },
        {   3,  69, 109 }, {   1,  44,  70 }, {   1,  18,  29 }
      }, {  // Band 3
        {  57, 186, 211 }, {  30, 140, 196 }, {   4,  93, 146 },
        {   1,  62, 102 }, {   1,  38,  65 }, {   1,  16,  27 }
      }, {  // Band 4
        {  47, 199, 217 }, {  14, 145, 196 }, {   1,  88, 142 },
        {   1,  57,  98 }, {   1,  36,  62 }, {   1,  15,  26 }
      }, {  // Band 5
        {  26, 219, 229 }, {   5, 155, 207 }, {   1,  94, 151 },
        {   1,  60, 104 }, {   1,  36,  62 }, {   1,  16,  28 }
      }
    }, {  // Inter
      {  // Band 0
        { 233,  29, 248 }, { 146,  47, 220 }, {  43,  52, 140 }
      }, {  // Band 1
        { 100, 163, 232 }, { 179, 161, 222 }, {  63, 142, 204 },
        {  37, 113, 174 }, {  26,  89, 137 }, {  18,  68,  97 }
      }, {  // Band 2
        {  85, 181, 230 }, {  32, 146, 209 }, {   7, 100, 164 },
        {   3,  71, 121 }, {   1,  45,  77 }, {   1,  18,  30 }
      }, {  // Band 3
        {  65, 187, 230 }, {  20, 148, 207 }, {   2,  97, 159 },
        {   1,  68, 116 }, {   1,  40,  70 }, {   1,  14,  29 }
      }, {  // Band 4
        {  40, 194, 227 }, {   8, 147, 204 }, {   1,  94, 155 },
        {   1,  65, 112 }, {   1,  39,  66 }, {   1,  14,  26 }
      }, {  // Band 5
        {  16, 208, 228 }, {   3, 151, 207 }, {   1,  98, 160 },
        {   1,  67, 117 }, {   1,  41,  74 }, {   1,  17,  31 }
      }
    }
  }
};

static const av1_coeff_probs_model default_coef_probs_32x32[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {  17,  38, 140 }, {   7,  34,  80 }, {   1,  17,  29 }
      }, {  // Band 1
        {  37,  75, 128 }, {  41,  76, 128 }, {  26,  66, 116 },
        {  12,  52,  94 }, {   2,  32,  55 }, {   1,  10,  16 }
      }, {  // Band 2
        {  50, 127, 154 }, {  37, 109, 152 }, {  16,  82, 121 },
        {   5,  59,  85 }, {   1,  35,  54 }, {   1,  13,  20 }
      }, {  // Band 3
        {  40, 142, 167 }, {  17, 110, 157 }, {   2,  71, 112 },
        {   1,  44,  72 }, {   1,  27,  45 }, {   1,  11,  17 }
      }, {  // Band 4
        {  30, 175, 188 }, {   9, 124, 169 }, {   1,  74, 116 },
        {   1,  48,  78 }, {   1,  30,  49 }, {   1,  11,  18 }
      }, {  // Band 5
        {  10, 222, 223 }, {   2, 150, 194 }, {   1,  83, 128 },
        {   1,  48,  79 }, {   1,  27,  45 }, {   1,  11,  17 }
      }
    }, {  // Inter
      {  // Band 0
        {  36,  41, 235 }, {  29,  36, 193 }, {  10,  27, 111 }
      }, {  // Band 1
        {  85, 165, 222 }, { 177, 162, 215 }, { 110, 135, 195 },
        {  57, 113, 168 }, {  23,  83, 120 }, {  10,  49,  61 }
      }, {  // Band 2
        {  85, 190, 223 }, {  36, 139, 200 }, {   5,  90, 146 },
        {   1,  60, 103 }, {   1,  38,  65 }, {   1,  18,  30 }
      }, {  // Band 3
        {  72, 202, 223 }, {  23, 141, 199 }, {   2,  86, 140 },
        {   1,  56,  97 }, {   1,  36,  61 }, {   1,  16,  27 }
      }, {  // Band 4
        {  55, 218, 225 }, {  13, 145, 200 }, {   1,  86, 141 },
        {   1,  57,  99 }, {   1,  35,  61 }, {   1,  13,  22 }
      }, {  // Band 5
        {  15, 235, 212 }, {   1, 132, 184 }, {   1,  84, 139 },
        {   1,  57,  97 }, {   1,  34,  56 }, {   1,  14,  23 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 181,  21, 201 }, {  61,  37, 123 }, {  10,  38,  71 }
      }, {  // Band 1
        {  47, 106, 172 }, {  95, 104, 173 }, {  42,  93, 159 },
        {  18,  77, 131 }, {   4,  50,  81 }, {   1,  17,  23 }
      }, {  // Band 2
        {  62, 147, 199 }, {  44, 130, 189 }, {  28, 102, 154 },
        {  18,  75, 115 }, {   2,  44,  65 }, {   1,  12,  19 }
      }, {  // Band 3
        {  55, 153, 210 }, {  24, 130, 194 }, {   3,  93, 146 },
        {   1,  61,  97 }, {   1,  31,  50 }, {   1,  10,  16 }
      }, {  // Band 4
        {  49, 186, 223 }, {  17, 148, 204 }, {   1,  96, 142 },
        {   1,  53,  83 }, {   1,  26,  44 }, {   1,  11,  17 }
      }, {  // Band 5
        {  13, 217, 212 }, {   2, 136, 180 }, {   1,  78, 124 },
        {   1,  50,  83 }, {   1,  29,  49 }, {   1,  14,  23 }
      }
    }, {  // Inter
      {  // Band 0
        { 197,  13, 247 }, {  82,  17, 222 }, {  25,  17, 162 }
      }, {  // Band 1
        { 126, 186, 247 }, { 234, 191, 243 }, { 176, 177, 234 },
        { 104, 158, 220 }, {  66, 128, 186 }, {  55,  90, 137 }
      }, {  // Band 2
        { 111, 197, 242 }, {  46, 158, 219 }, {   9, 104, 171 },
        {   2,  65, 125 }, {   1,  44,  80 }, {   1,  17,  91 }
      }, {  // Band 3
        { 104, 208, 245 }, {  39, 168, 224 }, {   3, 109, 162 },
        {   1,  79, 124 }, {   1,  50, 102 }, {   1,  43, 102 }
      }, {  // Band 4
        {  84, 220, 246 }, {  31, 177, 231 }, {   2, 115, 180 },
        {   1,  79, 134 }, {   1,  55,  77 }, {   1,  60,  79 }
      }, {  // Band 5
        {  43, 243, 240 }, {   8, 180, 217 }, {   1, 115, 166 },
        {   1,  84, 121 }, {   1,  51,  67 }, {   1,  16,   6 }
      }
    }
  }
};

#if CONFIG_TX64X64
static const av1_coeff_probs_model default_coef_probs_64x64[PLANE_TYPES] = {
  {  // Y plane
    {  // Intra
      {  // Band 0
        {  17,  38, 140 }, {   7,  34,  80 }, {   1,  17,  29 }
      }, {  // Band 1
        {  37,  75, 128 }, {  41,  76, 128 }, {  26,  66, 116 },
        {  12,  52,  94 }, {   2,  32,  55 }, {   1,  10,  16 }
      }, {  // Band 2
        {  50, 127, 154 }, {  37, 109, 152 }, {  16,  82, 121 },
        {   5,  59,  85 }, {   1,  35,  54 }, {   1,  13,  20 }
      }, {  // Band 3
        {  40, 142, 167 }, {  17, 110, 157 }, {   2,  71, 112 },
        {   1,  44,  72 }, {   1,  27,  45 }, {   1,  11,  17 }
      }, {  // Band 4
        {  30, 175, 188 }, {   9, 124, 169 }, {   1,  74, 116 },
        {   1,  48,  78 }, {   1,  30,  49 }, {   1,  11,  18 }
      }, {  // Band 5
        {  10, 222, 223 }, {   2, 150, 194 }, {   1,  83, 128 },
        {   1,  48,  79 }, {   1,  27,  45 }, {   1,  11,  17 }
      }
    }, {  // Inter
      {  // Band 0
        {  36,  41, 235 }, {  29,  36, 193 }, {  10,  27, 111 }
      }, {  // Band 1
        {  85, 165, 222 }, { 177, 162, 215 }, { 110, 135, 195 },
        {  57, 113, 168 }, {  23,  83, 120 }, {  10,  49,  61 }
      }, {  // Band 2
        {  85, 190, 223 }, {  36, 139, 200 }, {   5,  90, 146 },
        {   1,  60, 103 }, {   1,  38,  65 }, {   1,  18,  30 }
      }, {  // Band 3
        {  72, 202, 223 }, {  23, 141, 199 }, {   2,  86, 140 },
        {   1,  56,  97 }, {   1,  36,  61 }, {   1,  16,  27 }
      }, {  // Band 4
        {  55, 218, 225 }, {  13, 145, 200 }, {   1,  86, 141 },
        {   1,  57,  99 }, {   1,  35,  61 }, {   1,  13,  22 }
      }, {  // Band 5
        {  15, 235, 212 }, {   1, 132, 184 }, {   1,  84, 139 },
        {   1,  57,  97 }, {   1,  34,  56 }, {   1,  14,  23 }
      }
    }
  }, {  // UV plane
    {  // Intra
      {  // Band 0
        { 181,  21, 201 }, {  61,  37, 123 }, {  10,  38,  71 }
      }, {  // Band 1
        {  47, 106, 172 }, {  95, 104, 173 }, {  42,  93, 159 },
        {  18,  77, 131 }, {   4,  50,  81 }, {   1,  17,  23 }
      }, {  // Band 2
        {  62, 147, 199 }, {  44, 130, 189 }, {  28, 102, 154 },
        {  18,  75, 115 }, {   2,  44,  65 }, {   1,  12,  19 }
      }, {  // Band 3
        {  55, 153, 210 }, {  24, 130, 194 }, {   3,  93, 146 },
        {   1,  61,  97 }, {   1,  31,  50 }, {   1,  10,  16 }
      }, {  // Band 4
        {  49, 186, 223 }, {  17, 148, 204 }, {   1,  96, 142 },
        {   1,  53,  83 }, {   1,  26,  44 }, {   1,  11,  17 }
      }, {  // Band 5
        {  13, 217, 212 }, {   2, 136, 180 }, {   1,  78, 124 },
        {   1,  50,  83 }, {   1,  29,  49 }, {   1,  14,  23 }
      }
    }, {  // Inter
      {  // Band 0
        { 197,  13, 247 }, {  82,  17, 222 }, {  25,  17, 162 }
      }, {  // Band 1
        { 126, 186, 247 }, { 234, 191, 243 }, { 176, 177, 234 },
        { 104, 158, 220 }, {  66, 128, 186 }, {  55,  90, 137 }
      }, {  // Band 2
        { 111, 197, 242 }, {  46, 158, 219 }, {   9, 104, 171 },
        {   2,  65, 125 }, {   1,  44,  80 }, {   1,  17,  91 }
      }, {  // Band 3
        { 104, 208, 245 }, {  39, 168, 224 }, {   3, 109, 162 },
        {   1,  79, 124 }, {   1,  50, 102 }, {   1,  43, 102 }
      }, {  // Band 4
        {  84, 220, 246 }, {  31, 177, 231 }, {   2, 115, 180 },
        {   1,  79, 134 }, {   1,  55,  77 }, {   1,  60,  79 }
      }, {  // Band 5
        {  43, 243, 240 }, {   8, 180, 217 }, {   1, 115, 166 },
        {   1,  84, 121 }, {   1,  51,  67 }, {   1,  16,   6 }
      }
    }
  }
};
#endif  // CONFIG_TX64X64
#endif  // CONFIG_ENTROPY
/* clang-format on */

static void extend_to_full_distribution(aom_prob *probs, aom_prob p) {
  assert(p != 0);
  memcpy(probs, av1_pareto8_full[p - 1], MODEL_NODES * sizeof(aom_prob));
}

void av1_model_to_full_probs(const aom_prob *model, aom_prob *full) {
  if (full != model)
    memcpy(full, model, sizeof(aom_prob) * UNCONSTRAINED_NODES);
  extend_to_full_distribution(&full[UNCONSTRAINED_NODES], model[PIVOT_NODE]);
}

#if CONFIG_EC_MULTISYMBOL
static void build_token_cdfs(const aom_prob *pdf_model,
                             aom_cdf_prob cdf[ENTROPY_TOKENS]) {
  int i, sum = 0;
  assert(pdf_model[2] != 0);
  for (i = 0; i < ENTROPY_TOKENS - 2; ++i) {
    cdf[i] = sum += av1_pareto8_token_probs[pdf_model[2] - 1][i];
  }
}

void av1_coef_pareto_cdfs(FRAME_CONTEXT *fc) {
  TX_SIZE t;
  int i, j, k, l;
  for (t = TX_4X4; t < TX_SIZES; ++t)
    for (i = 0; i < PLANE_TYPES; ++i)
      for (j = 0; j < REF_TYPES; ++j)
        for (k = 0; k < COEF_BANDS; ++k)
          for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l)
            build_token_cdfs(fc->coef_probs[t][i][j][k][l],
                             fc->coef_cdfs[t][i][j][k][l]);
}
#endif  // CONFIG_EC_MULTISYMBOL

void av1_default_coef_probs(AV1_COMMON *cm) {
#if CONFIG_ENTROPY
  const int index = AOMMIN(
      ROUND_POWER_OF_TWO(cm->base_qindex, 8 - QCTX_BIN_BITS), QCTX_BINS - 1);
  av1_copy(cm->fc->coef_probs, default_qctx_coef_probs[index]);
#else
#if CONFIG_CB4X4
  av1_copy(cm->fc->coef_probs[TX_2X2], default_coef_probs_4x4);
#endif
  av1_copy(cm->fc->coef_probs[TX_4X4], default_coef_probs_4x4);
  av1_copy(cm->fc->coef_probs[TX_8X8], default_coef_probs_8x8);
  av1_copy(cm->fc->coef_probs[TX_16X16], default_coef_probs_16x16);
  av1_copy(cm->fc->coef_probs[TX_32X32], default_coef_probs_32x32);
#if CONFIG_TX64X64
  av1_copy(cm->fc->coef_probs[TX_64X64], default_coef_probs_64x64);
#endif  // CONFIG_TX64X64
#endif  // CONFIG_ENTROPY
#if CONFIG_EC_MULTISYMBOL
  av1_coef_pareto_cdfs(cm->fc);
#endif  // CONFIG_EC_MULTISYMBOL
}

static void adapt_coef_probs(AV1_COMMON *cm, TX_SIZE tx_size,
                             unsigned int count_sat,
                             unsigned int update_factor) {
  const FRAME_CONTEXT *pre_fc = &cm->frame_contexts[cm->frame_context_idx];
  av1_coeff_probs_model *const probs = cm->fc->coef_probs[tx_size];
#if CONFIG_ENTROPY
  const av1_coeff_probs_model *const pre_probs =
      cm->partial_prob_update
          ? (const av1_coeff_probs_model *)cm->starting_coef_probs[tx_size]
          : pre_fc->coef_probs[tx_size];
#else
  const av1_coeff_probs_model *const pre_probs = pre_fc->coef_probs[tx_size];
#endif  // CONFIG_ENTROPY
  const av1_coeff_count_model *const counts =
      (const av1_coeff_count_model *)cm->counts.coef[tx_size];
  const unsigned int(*eob_counts)[REF_TYPES][COEF_BANDS][COEFF_CONTEXTS] =
      (const unsigned int(*)[REF_TYPES][COEF_BANDS]
                            [COEFF_CONTEXTS])cm->counts.eob_branch[tx_size];
  int i, j, k, l, m;

  for (i = 0; i < PLANE_TYPES; ++i)
    for (j = 0; j < REF_TYPES; ++j)
      for (k = 0; k < COEF_BANDS; ++k)
        for (l = 0; l < BAND_COEFF_CONTEXTS(k); ++l) {
          const int n0 = counts[i][j][k][l][ZERO_TOKEN];
          const int n1 = counts[i][j][k][l][ONE_TOKEN];
          const int n2 = counts[i][j][k][l][TWO_TOKEN];
          const int neob = counts[i][j][k][l][EOB_MODEL_TOKEN];
          const unsigned int branch_ct[UNCONSTRAINED_NODES][2] = {
            { neob, eob_counts[i][j][k][l] - neob }, { n0, n1 + n2 }, { n1, n2 }
          };
          for (m = 0; m < UNCONSTRAINED_NODES; ++m)
            probs[i][j][k][l][m] =
                av1_merge_probs(pre_probs[i][j][k][l][m], branch_ct[m],
                                count_sat, update_factor);
        }
}

void av1_adapt_coef_probs(AV1_COMMON *cm) {
  TX_SIZE tx_size;
  unsigned int count_sat, update_factor;

#if CONFIG_ADAPT_SCAN
  TX_TYPE tx_type;
#endif

#if CONFIG_ENTROPY
  if (!frame_is_intra_only(cm) && cm->last_frame_type == KEY_FRAME) {
    update_factor = COEF_MAX_UPDATE_FACTOR_AFTER_KEY_BITS; /* adapt quickly */
    count_sat = COEF_COUNT_SAT_AFTER_KEY_BITS;
  } else {
    update_factor = COEF_MAX_UPDATE_FACTOR_BITS;
    count_sat = COEF_COUNT_SAT_BITS;
  }
  if (cm->partial_prob_update == 1) {
    update_factor = COEF_MAX_UPDATE_FACTOR_BITS;
  }
#else
  if (!frame_is_intra_only(cm) && cm->last_frame_type == KEY_FRAME) {
    update_factor = COEF_MAX_UPDATE_FACTOR_AFTER_KEY; /* adapt quickly */
    count_sat = COEF_COUNT_SAT_AFTER_KEY;
  } else {
    update_factor = COEF_MAX_UPDATE_FACTOR;
    count_sat = COEF_COUNT_SAT;
  }
#endif  // CONFIG_ENTROPY
  for (tx_size = 0; tx_size < TX_SIZES; tx_size++)
    adapt_coef_probs(cm, tx_size, count_sat, update_factor);

#if CONFIG_ADAPT_SCAN
  for (tx_size = TX_4X4; tx_size < TX_SIZES; ++tx_size)
    for (tx_type = DCT_DCT; tx_type < TX_TYPES; ++tx_type) {
      av1_update_scan_prob(cm, tx_size, tx_type, ADAPT_SCAN_UPDATE_RATE_16);
      av1_update_scan_order_facade(cm, tx_size, tx_type);
    }
#endif
}

#if CONFIG_ENTROPY
void av1_partial_adapt_probs(AV1_COMMON *cm, int mi_row, int mi_col) {
  (void)mi_row;
  (void)mi_col;

  if (cm->refresh_frame_context == REFRESH_FRAME_CONTEXT_BACKWARD) {
    cm->partial_prob_update = 1;
    av1_adapt_coef_probs(cm);
  }
}
#endif  // CONFIG_ENTROPY
